from collections.abc import Iterable


def print_args_and_kwargs(args: Iterable, kwargs: dict) -> None:
    if args:
        print('args:')
        for i, arg in enumerate(args):
            arg = arg_to_str(arg)
            print(f'  {i}: {arg}')

    if kwargs:
        print('kwargs:')
        for key, value in kwargs.items():
            value = arg_to_str(value)
            print(f'  {key}={value}')


def arg_to_str(arg) -> str:
    arg = repr(arg)

    if '\n' in arg:
        arg = f'```\n{arg}\n```'

    return arg



import asyncio
import hmac
import secrets
from abc import ABC, abstractmethod
from asyncio import IncompleteReadError
from collections.abc import Callable

from rosy.asyncio import Reader, Writer
from rosy.utils import require

AuthKey = bytes


class Authenticator(ABC):
    @abstractmethod
    async def authenticate(self, reader: Reader, writer: Writer) -> None:
        """Raises AuthenticationError if authentication fails."""
        ...


class NoAuthenticator(Authenticator):
    """Performs no authentication."""

    async def authenticate(self, reader: Reader, writer: Writer) -> None:
        pass


class HMACAuthenticator(Authenticator):
    """
    Authenticates using symmetric HMAC challenge-response with a shared secret key.
    """

    def __init__(
            self,
            authkey: AuthKey,
            challenge_length: int = 32,
            digest='sha256',
            timeout: float | None = 10.,
            get_random_bytes: Callable[[int], bytes] = secrets.token_bytes,
    ):
        require(authkey, 'authkey must not be empty.')
        require(challenge_length > 0, 'challenge_length must be > 0.')

        self.authkey = authkey
        self.challenge_length = challenge_length
        self.digest = digest
        self.timeout = timeout
        self.get_random_bytes = get_random_bytes

    async def authenticate(self, reader: Reader, writer: Writer) -> None:
        challenge_to_client = self.get_random_bytes(self.challenge_length)

        writer.write(challenge_to_client)
        await writer.drain()

        challenge_from_client = await self._read_exactly(reader, self.challenge_length)

        hmac_to_client = self._hmac_digest(challenge_from_client)

        writer.write(hmac_to_client)
        await writer.drain()

        expected_hmac = self._hmac_digest(challenge_to_client)

        hmac_from_client = await self._read_exactly(reader, len(expected_hmac))

        if not hmac.compare_digest(hmac_from_client, expected_hmac):
            raise AuthenticationError('Received HMAC digest does not match expected digest.')

    async def _read_exactly(self, reader: Reader, n: int) -> bytes:
        try:
            return await asyncio.wait_for(
                reader.readexactly(n),
                timeout=self.timeout,
            )
        except TimeoutError:
            raise AuthenticationError(f'Timeout after {self.timeout}s waiting for authkey.')
        except IncompleteReadError as e:
            raise AuthenticationError(e)

    def _hmac_digest(self, challenge: bytes) -> bytes:
        return hmac.digest(self.authkey, challenge, self.digest)


class AuthenticationError(Exception):
    pass


def optional_authkey_authenticator(
        authkey: AuthKey | None,
) -> HMACAuthenticator | NoAuthenticator:
    return HMACAuthenticator(authkey) if authkey else NoAuthenticator()



from argparse import ArgumentParser, Namespace

from rosy.cli.topic.echo import add_echo_command, echo_main
from rosy.cli.topic.list import add_list_command, list_main
from rosy.cli.topic.send import add_send_command, send_main


async def topic_main(args: Namespace) -> None:
    if args.topic_command == 'list':
        await list_main(args)
    elif args.topic_command == 'echo':
        await echo_main(args)
    elif args.topic_command == 'send':
        await send_main(args)
    else:
        raise ValueError(f'Unknown topic command: {args.topic_command}')


def add_topic_command(subparsers) -> None:
    parser: ArgumentParser = subparsers.add_parser(
        'topic',
        description='Topic commands.',
        help='Topic commands like send, echo, etc.',
    )

    subparsers = parser.add_subparsers(
        title='commands',
        dest='topic_command',
        required=True,
    )

    add_list_command(subparsers)
    add_echo_command(subparsers)
    add_send_command(subparsers)



import signal
from dataclasses import dataclass
from subprocess import Popen, TimeoutExpired
from typing import Any


class ProcessManager:
    processes: list[Popen]
    """Processes managed by this manager."""

    timeout: float | None
    """Default timeout for stopping processes."""

    def __init__(
            self,
            python_exe: str = 'python',
            options: dict[str, Any] = None,
            timeout: float | None = 10.,
    ):
        """
        ProcessManager makes it easy to start and stop numerous processes
        in a coordinated fashion.

        Example:

            with ProcessManager() as pm:
                # Run: ls -l
                pm.popen(['ls', '-l'])

                pm.start_python('my_script.py', arg1='value1')
                pm.start_python('my_other_script.py', arg2='value2')

        Args:
            python_exe:
                Path to the Python executable to use for starting Python processes.
            options:
                Default kwargs to pass to Popen.
            timeout:
                Default timeout for stopping processes.
        """

        self.python_exe = python_exe
        self.timeout = timeout

        self.processes: list[Popen] = []
        self._options: dict[str, Any] = options or {}

    @property
    def options(self) -> dict[str, Any]:
        return dict(self._options)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()

    def using_options(self, **options) -> '_OptionsContext':
        """
        Use with a ``with`` statement to temporarily use different default
        options for Popen. Useful when you want to run a bunch of processes
        with a different set of options.

        Example:

            with ProcessManager(shell=False) as pm:
                pm.popen(['ls', '-l'])

                with pm.using_options(shell=True):
                    pm.popen('ls -l')
                    pm.popen('cp foo bar')
                    pm.popen('cat bar')

        Args:
            **options:
                New default options to use for Popen.

        Returns: A context manager.
        """

        return _OptionsContext(self, options)

    def popen(self, args, **kwargs) -> Popen:
        """
        Start a new process.

        Args:
            args: Command arguments passed to Popen.
            **kwargs: Keyword arguments passed to Popen.

        Returns: The Popen object for the new process.
        """

        kwargs = {**self._options, **kwargs}
        return self.add(Popen(args, **kwargs))

    def add(self, process: Popen) -> Popen:
        """Add an existing process to the manager."""
        self.processes.append(process)
        return process

    def start_python(self, args, **kwargs) -> Popen:
        """
        Start a new Python process.

        Args:
            args: Command arguments appended to the Python exe and passed to Popen.
            **kwargs: Keyword arguments passed to Popen.

        Returns: The Popen object for the new process.
        """

        if isinstance(args, str):
            args = f'"{self.python_exe}" {args}'
        else:
            args = [self.python_exe] + list(args)

        return self.popen(args, **kwargs)

    def stop(
            self,
            process: Popen | None = None,
            timeout: float | None = None,
    ):
        """
        Stop one or all processes managed by this manager.

        First it sends a `SIGTERM` signal to the process, and then waits for it to
        finish. If the process does not finish within the timeout, it sends a
        `SIGKILL` signal to forcefully terminate the process.

        Args:
            process:
                Optional process to stop. By default, all processes will be stopped.
            timeout:
                Optional timeout override. Defaults to `self.timeout`.
        """

        if process is None:
            processes = list(self.processes)
            self.processes.clear()
        else:
            processes = [process]
            self.processes.remove(process)

        if timeout is None:
            timeout = self.timeout

        print(f'Stopping {len(processes)} processes...')
        for process in processes:
            process.send_signal(signal.SIGTERM)

        procs_to_kill = []
        for process in processes:
            try:
                process.wait(timeout)
            except TimeoutExpired:
                print(f'Process {process.pid} did not stop in time.')
                procs_to_kill.append(process)

        for process in procs_to_kill:
            print(f'Killing process {process.pid}')
            process.kill()

        for process in procs_to_kill:
            try:
                process.wait(timeout)
            except TimeoutExpired:
                print(f'Failed to kill process {process.pid}.')

    def wait(self, timeout: float = None) -> None:
        """
        Wait for all processes to finish.

        Args:
            timeout: Optional wait timeout.
        """

        for process in list(self.processes):
            process.wait(timeout)
            self.processes.remove(process)


@dataclass
class _OptionsContext:
    procman: ProcessManager
    options: dict[str, Any]
    prev_options: dict[str, Any] = None

    def __enter__(self):
        self.prev_options = self.procman._options
        self.procman._options = self.options
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.procman._options = self.prev_options



import logging
from argparse import ArgumentParser, Namespace

from rosy.argparse import add_authkey_arg, add_coordinator_arg
from rosy.cli.utils import add_log_arg, get_mesh_topology


async def list_main(args: Namespace):
    logging.basicConfig(level=args.log)

    topology = await get_mesh_topology(args)

    topics = sorted({
        topic
        for node in topology.nodes
        for topic in node.topics
    })

    for topic in topics:
        print(topic)


def add_list_command(subparsers) -> None:
    parser: ArgumentParser = subparsers.add_parser(
        'list',
        description='List all topics currently being listened to by nodes.',
        help='list topics being listened to',
    )

    add_log_arg(parser)

    add_coordinator_arg(parser)
    add_authkey_arg(parser)



from rosy.asyncio import Reader, Writer
from rosy.codec import Codec
from rosy.node.service.types import RequestId, ServiceRequest, ServiceResponse
from rosy.node.types import Args, KWArgs
from rosy.types import Data, Service


class ServiceRequestCodec(Codec[ServiceRequest]):
    def __init__(
            self,
            id_codec: Codec[RequestId],
            service_codec: Codec[Service],
            args_codec: Codec[Args],
            kwargs_codec: Codec[KWArgs],
    ):
        self.id_codec = id_codec
        self.service_codec = service_codec
        self.args_codec = args_codec
        self.kwargs_codec = kwargs_codec

    async def encode(self, writer: Writer, request: ServiceRequest) -> None:
        await self.id_codec.encode(writer, request.id)
        await self.service_codec.encode(writer, request.service)
        await self.args_codec.encode(writer, request.args)
        await self.kwargs_codec.encode(writer, request.kwargs)

    async def decode(self, reader: Reader) -> ServiceRequest:
        id = await self.id_codec.decode(reader)
        service = await self.service_codec.decode(reader)
        args = await self.args_codec.decode(reader)
        kwargs = await self.kwargs_codec.decode(reader)
        return ServiceRequest(id, service, args, kwargs)


class ServiceResponseCodec(Codec[ServiceResponse]):
    def __init__(
            self,
            id_codec: Codec[RequestId],
            data_codec: Codec[Data],
            error_codec: Codec[str],
            success_status_code: bytes = b'\x00',
            error_status_code: bytes = b'\xEE',
    ):
        self.id_codec = id_codec
        self.data_codec = data_codec
        self.error_codec = error_codec
        self.success_status_code = success_status_code
        self.error_status_code = error_status_code

    async def encode(self, writer: Writer, response: ServiceResponse) -> None:
        await self.id_codec.encode(writer, response.id)

        if response.error:
            writer.write(self.error_status_code)
            await self.error_codec.encode(writer, response.error)
        else:
            writer.write(self.success_status_code)
            await self.data_codec.encode(writer, response.result)

    async def decode(self, reader: Reader) -> ServiceResponse:
        id = await self.id_codec.decode(reader)

        status_code = await reader.readexactly(1)
        if status_code == self.success_status_code:
            data = await self.data_codec.decode(reader)
            error = None
        elif status_code == self.error_status_code:
            data = None
            error = await self.error_codec.decode(reader)
        else:
            raise ValueError(f'Received unknown status code={status_code!r}')

        return ServiceResponse(id, data, error)



from argparse import Namespace
from collections import Counter
from dataclasses import dataclass
from datetime import datetime, timedelta
from pathlib import Path

from rosy.cli.bag.file import get_bag_file_messages, get_most_recent_bag_file_path


async def display_info(args: Namespace) -> None:
    bag_file_path = args.input or get_most_recent_bag_file_path()

    info = get_info(bag_file_path)

    print(f'''
path:     {info.path}
duration: {info.duration}
start:    {info.start}
end:      {info.end}
size:     {info.size} {info.size_unit}
messages: {info.messages}
'''.strip())

    if info.topics:
        print('topics:')
        for topic, count in info.topics.items():
            pct = round(100 * count / info.messages)
            print(f'- {topic!r}:\t{count} ({pct}%)')
    else:
        print('topics:   None')


def add_info_args(subparsers) -> None:
    parser = subparsers.add_parser('info', help='Get info about recorded messages in a file')

    parser.add_argument(
        '--input', '-i',
        type=Path,
        help='Input file path. Default: The most recent '
             'record_*.bag file in the current directory.',
    )


def get_info(bag_file_path: Path) -> 'BagInfo':
    size = bag_file_path.stat().st_size
    size, size_unit = get_human_readable_size(size)

    start_time = end_time = None
    messages = 0
    topics = Counter()

    for instant, topic, *_ in get_bag_file_messages(bag_file_path):
        if not start_time:
            start_time = instant

        end_time = instant
        messages += 1
        topics[topic] += 1

    return BagInfo(
        bag_file_path,
        start_time,
        end_time,
        size,
        size_unit,
        messages,
        topics,
    )


def get_human_readable_size(size: int) -> tuple[int, str]:
    for unit in ['B', 'KB', 'MB']:
        if size < 1024:
            return round(size), unit
        size /= 1024

    return round(size), 'GB'



@dataclass
class BagInfo:
    path: Path
    start: datetime
    end: datetime
    size: int
    size_unit: str
    messages: int
    topics: dict[str, int]

    @property
    def duration(self) -> timedelta:
        start, end = self.start, self.end
        return (end - start) if start and end else None



from rosy.asyncio import Reader, Writer
from rosy.codec import Codec
from rosy.node.topic.types import TopicMessage
from rosy.node.types import Args, KWArgs
from rosy.types import Topic


class TopicMessageCodec(Codec[TopicMessage]):
    def __init__(
            self,
            topic_codec: Codec[Topic],
            args_codec: Codec[Args],
            kwargs_codec: Codec[KWArgs],
    ):
        self.topic_codec = topic_codec
        self.args_codec = args_codec
        self.kwargs_codec = kwargs_codec

    async def encode(self, writer: Writer, message: TopicMessage) -> None:
        await self.topic_codec.encode(writer, message.topic)
        await self.args_codec.encode(writer, message.args)
        await self.kwargs_codec.encode(writer, message.kwargs)

    async def decode(self, reader: Reader) -> TopicMessage:
        topic = await self.topic_codec.decode(reader)
        args = await self.args_codec.decode(reader)
        kwargs = await self.kwargs_codec.decode(reader)
        return TopicMessage(topic, args, kwargs)



from argparse import ArgumentParser
from collections.abc import Callable
from urllib.parse import urlparse

from rosy.coordinator.constants import DEFAULT_COORDINATOR_PORT
from rosy.types import Endpoint, Port


def get_node_arg_parser(
        default_node_name: str = None,
        default_coordinator: Endpoint = None,
        default_authkey: bytes = None,
        **kwargs,
) -> ArgumentParser:
    """
    Returns an argument parser with the following node-specific arguments:
    - `--name`
    - `--coordinator HOST[:PORT]`
    - `--authkey`

    Args:
        default_node_name:
            Default node name. If not given, the argument `--name` is required.
        default_coordinator:
            Default coordinator endpoint. If not given, the argument will
            default to `localhost:DEFAULT_COORDINATOR_PORT`.
        default_authkey:
            Default authentication key. If not given, the argument will
            default to `None` (i.e. no auth key).
        **kwargs:
            These are passed to the `ArgumentParser` constructor.
    """

    parser = ArgumentParser(**kwargs)
    add_node_args(parser, default_node_name, default_coordinator, default_authkey)
    return parser


def add_node_args(
        parser: ArgumentParser,
        default_node_name: str = None,
        default_coordinator: Endpoint = None,
        default_authkey: bytes = None,
) -> None:
    """
    Adds node-specific arguments to an argument parser:
    - `--name`
    - `--coordinator HOST[:PORT]`
    - `--authkey`

    Args:
        parser:
            Argument parser to which the arguments will be added.
        default_node_name:
            Default node name. If not given, the argument is required.
        default_coordinator:
            Default coordinator endpoint. If not given, the argument will
            default to `localhost:DEFAULT_COORDINATOR_PORT`.
        default_authkey:
            Default authentication key. If not given, the argument will
            default to `None` (i.e. no auth key).
    """

    add_node_name_arg(parser, default_node_name)
    add_coordinator_arg(parser, default_coordinator)
    add_authkey_arg(parser, default_authkey)


def add_node_name_arg(parser: ArgumentParser, default: str = None) -> None:
    """
    Adds a `--name` argument to an argument parser.

    Args:
        parser:
            Argument parser to which the argument will be added.
        default:
            Default node name. If not given, the argument will be required.
    """

    arg_args = dict(
        default=default,
        help='Node name. Default: %(default)s',
    ) if default is not None else dict(
        required=True,
        help='Node name.',
    )

    parser.add_argument('--name', **arg_args)


def add_coordinator_arg(
        parser: ArgumentParser,
        default: Endpoint = None,
) -> None:
    """
    Adds a `--coordinator HOST[:PORT]` argument to an argument parser.

    Args:
        parser:
            Argument parser to which the argument will be added.
        default:
            Default coordinator endpoint. If not given, the argument will
            default to `localhost:DEFAULT_COORDINATOR_PORT`.
    """

    if default is None:
        default = Endpoint('localhost', DEFAULT_COORDINATOR_PORT)

    parser.add_argument(
        '--coordinator',
        default=default,
        type=endpoint_arg(default_port=default.port),
        metavar='HOST[:PORT]',
        help=f'Coordinator host and port. Default: %(default)s',
    )


def add_authkey_arg(
        parser: ArgumentParser,
        default: bytes = None,
) -> None:
    """
    Adds an `--authkey` argument to an argument parser.

    Args:
        parser:
            Argument parser to which the argument will be added.
        default:
            Default authentication key. If not given, the argument will
            default to `None` (i.e. no auth key).
    """

    parser.add_argument(
        '--authkey',
        default=default,
        type=lambda arg: arg.encode(),
        help='Authentication key to use for new connections between all nodes in the mesh. '
             'Should ideally be >= 32 characters. Default: %(default)s',
    )


def endpoint_arg(default_port: Port = None) -> Callable[[str], Endpoint]:
    """
    Returns a function that parses a string of format ``host[:port]`` into an
    ``Endpoint`` object. If the port is not specified, it defaults to the
    provided ``default_port``.
    """

    def wrapped(arg: str) -> Endpoint:
        parsed = urlparse(f'//{arg}')
        host = parsed.hostname
        port = parsed.port or default_port
        return Endpoint(host, port)

    return wrapped



from typing import NamedTuple

from rosy.node.types import Args, KWArgs
from rosy.types import Data, Service

RequestId = int


class ServiceRequest(NamedTuple):
    id: RequestId
    service: Service
    args: Args
    kwargs: KWArgs


class ServiceResponse(NamedTuple):
    id: RequestId
    result: Data = None
    error: str | None = None



import asyncio
import logging
from abc import ABC, abstractmethod
from asyncio import Lock, open_connection, wait_for
from collections.abc import Awaitable, Callable

from rosy.authentication import Authenticator, NoAuthenticator
from rosy.coordinator.constants import DEFAULT_COORDINATOR_PORT
from rosy.objectio import CodecObjectReader, CodecObjectWriter, ObjectIO
from rosy.reqres import MeshTopologyBroadcast, RegisterNodeRequest, RegisterNodeResponse
from rosy.rpc import ObjectIORPC, RPC
from rosy.specs import MeshNodeSpec, MeshTopologySpec
from rosy.types import Host, Port

MeshTopologyBroadcastHandler = Callable[[MeshTopologyBroadcast], Awaitable[None]]

logger = logging.getLogger(__name__)


class MeshCoordinatorClient(ABC):
    mesh_topology_broadcast_handler: MeshTopologyBroadcastHandler

    def set_broadcast_handler(
            self, handler: MeshTopologyBroadcastHandler
    ) -> None:
        self.mesh_topology_broadcast_handler = handler

    @abstractmethod
    async def get_topology(self) -> MeshTopologySpec:
        ...

    @abstractmethod
    async def send_heartbeat(self) -> None:
        ...

    @abstractmethod
    async def register_node(self, node_spec: MeshNodeSpec) -> None:
        ...


class RPCMeshCoordinatorClient(MeshCoordinatorClient):
    def __init__(self, rpc: RPC):
        self.rpc = rpc
        rpc.message_handler = self._handle_rpc_message

    async def get_topology(self) -> MeshTopologySpec:
        response = await self.rpc.send_request(b'get_topology')

        if not isinstance(response, MeshTopologySpec):
            raise Exception(f'Failed to get topology; got response={response}')

        return response

    async def send_heartbeat(self) -> None:
        response = await self.rpc.send_request(b'ping')
        if response != b'pong':
            raise Exception(f'Got unexpected response={response} from heartbeat.')

    async def register_node(self, node_spec: MeshNodeSpec) -> None:
        request = RegisterNodeRequest(node_spec)

        response = await self.rpc.send_request(request)

        if not isinstance(response, RegisterNodeResponse):
            raise Exception(f'Failed to register node; got response={response}')

    async def _handle_rpc_message(self, data) -> None:
        if isinstance(data, MeshTopologyBroadcast):
            await self.mesh_topology_broadcast_handler(data)
        else:
            logger.error(f'Received unknown message={data}')


MeshCoordinatorClientBuilder = Callable[[], Awaitable[MeshCoordinatorClient]]


class ReconnectMeshCoordinatorClient(MeshCoordinatorClient):
    """
    Automatically reconnects to the coordinator when it detects that the
    connection has been lost.

    When a re-connection is made, the most recent node spec is re-registered
    with the coordinator, if applicable.
    """

    def __init__(
            self,
            client_builder: MeshCoordinatorClientBuilder,
            timeout: float,
    ):
        self.client_builder = client_builder
        self.timeout = timeout

        self._client: MeshCoordinatorClient | None = None
        self._client_lock = Lock()
        self._node_spec: MeshNodeSpec | None = None
        self._mesh_topology_broadcast_handler = None
        self._connection_monitor_task: asyncio.Task | None = None

    @property
    async def client(self) -> MeshCoordinatorClient:
        async with self._client_lock:
            if self._client is not None:
                return self._client

            try:
                self._client = await self.client_builder()
                self._client.mesh_topology_broadcast_handler = self.mesh_topology_broadcast_handler
            finally:
                if self._connection_monitor_task is None:
                    self._connection_monitor_task = asyncio.create_task(self._monitor_connection())

            return self._client

    @property
    def mesh_topology_broadcast_handler(self) -> MeshTopologyBroadcastHandler:
        return self._mesh_topology_broadcast_handler

    @mesh_topology_broadcast_handler.setter
    def mesh_topology_broadcast_handler(self, handler: MeshTopologyBroadcastHandler) -> None:
        self._mesh_topology_broadcast_handler = handler

        if self._client is not None:
            self._client.mesh_topology_broadcast_handler = handler

    async def get_topology(self) -> MeshTopologySpec:
        return await self._call_client_method('get_topology')

    async def send_heartbeat(self) -> None:
        await self._call_client_method('send_heartbeat')

    async def register_node(self, node_spec: MeshNodeSpec) -> None:
        self._node_spec = node_spec

        try:
            await self._call_client_method('register_node', node_spec)
        except Exception as e:
            logger.error(
                f'Failed to register node. Registration will be retried. {e!r}'
            )

    async def _call_client_method(self, method: str, *args, **kwargs):
        try:
            client = await self.client
            client_method = getattr(client, method)

            return await wait_for(
                client_method(*args, **kwargs),
                timeout=self.timeout,
            )
        except Exception:
            self._client = None
            raise

    async def _monitor_connection(self) -> None:
        prev_error = False
        while True:
            try:
                await self.send_heartbeat()
            except Exception as e:
                logger.error(f'Heartbeat failed: {e!r}')
                error = True
            else:
                error = False

            if prev_error and not error:
                logger.info('Connection to coordinator restored.')

                if self._node_spec:
                    await self.register_node(self._node_spec)

            await asyncio.sleep(self.timeout)
            prev_error = error


async def build_coordinator_client(
        host: Host,
        port: Port = DEFAULT_COORDINATOR_PORT,
        authenticator: Authenticator = None,
        reconnect_timeout: float | None = 5.0,
) -> MeshCoordinatorClient:
    if authenticator is None:
        authenticator = NoAuthenticator()

    async def client_builder():
        reader, writer = await open_connection(host, port)
        await authenticator.authenticate(reader, writer)

        rpc = ObjectIORPC(ObjectIO(
            reader=CodecObjectReader(reader),
            writer=CodecObjectWriter(writer),
        ))
        client_ = RPCMeshCoordinatorClient(rpc)
        await rpc.start()

        return client_

    if reconnect_timeout is not None and reconnect_timeout > 0:
        return ReconnectMeshCoordinatorClient(
            client_builder,
            timeout=reconnect_timeout,
        )
    else:
        return await client_builder()



import pickle
from datetime import datetime
from pathlib import Path
from typing import Iterable

from rosy.node.types import Args, KWArgs
from rosy.types import Topic

Message = tuple[datetime, Topic, Args, KWArgs]


def get_bag_file_messages(bag_file_path: Path) -> Iterable[Message]:
    with open(bag_file_path, 'rb') as bag_file:
        while True:
            try:
                yield pickle.load(bag_file)
            except EOFError:
                break


def get_most_recent_bag_file_path() -> Path:
    matching_files = Path('.').glob(
        'record_????-??-??-??-??-??.bag'
    )

    try:
        return max(matching_files)
    except ValueError:
        raise FileNotFoundError(
            'No bag files found in the current directory. '
            'Use --input to specify a file.'
        )



import logging
from argparse import ArgumentParser, Namespace

from rosy.argparse import add_authkey_arg
from rosy.asyncio import forever
from rosy.cli.utils import add_log_arg
from rosy.coordinator.constants import DEFAULT_COORDINATOR_HOST, DEFAULT_COORDINATOR_PORT
from rosy.coordinator.server import build_mesh_coordinator_server
from rosy.types import ServerHost


async def coordinator_main(args: Namespace) -> None:
    logging.basicConfig(level=args.log)

    server = build_mesh_coordinator_server(
        host=args.host,
        port=args.port,
        authkey=args.authkey,
    )

    try:
        await server.start()
    except:
        print(f'Failed to start rosy coordinator on {args.host}:{args.port}')
        raise
    else:
        print(f'Started rosy coordinator on {args.host}:{args.port}')

    await forever()


def add_coordinator_command(subparsers) -> None:
    parser: ArgumentParser = subparsers.add_parser(
        'coordinator',
        description='Start the rosy coordinator.',
        help='(Default) Start the rosy coordinator',
    )

    def server_host_arg(arg: str) -> ServerHost:
        if arg == 'None':
            return None

        hosts = arg.split(',')
        return hosts if len(hosts) > 1 else hosts[0]

    parser.add_argument(
        '--host', default=DEFAULT_COORDINATOR_HOST, type=server_host_arg,
        help='Comma-separated list of host(s) to bind to. '
             'Default is empty string, which means all available interfaces.',
    )

    parser.add_argument(
        '--port', default=DEFAULT_COORDINATOR_PORT, type=int,
        help=f'Port to bind to. Default is {DEFAULT_COORDINATOR_PORT}.',
    )

    add_authkey_arg(parser)

    parser.add_argument(
        '--log-heartbeats', action='store_true',
        help='Log heartbeats from nodes.',
    )

    add_log_arg(parser, default='INFO')



import asyncio
import traceback
from asyncio import Lock, Task
from collections.abc import Awaitable, Iterable, Sized
from io import BytesIO
from typing import Protocol, Type, TypeVar

from rosy.types import Buffer

T = TypeVar('T')
E = TypeVar('E', bound=BaseException)


async def close_ignoring_errors(writer: 'Writer') -> None:
    """Closes the writer ignoring any ConnectionErrors."""
    try:
        writer.close()
        await writer.wait_closed()
    except ConnectionError:
        pass


async def forever():
    """Never returns."""
    await asyncio.Future()


async def log_error(
        awaitable: Awaitable[T],
        base_exception: Type[E] = Exception,
) -> T | E:
    """
    Returns the result of the awaitable, logging any exceptions.

    If an exception occurs, it is returned.

    Args:
        awaitable: The awaitable to await.
        base_exception: The base type of exception to catch.
    """

    try:
        return await awaitable
    except base_exception as e:
        traceback.print_exc()
        return e


async def many(
        awaitables: Iterable[Awaitable[T]],
        base_exception: Type[E] = Exception,
) -> list[T | E]:
    """
    Await multiple awaitables in parallel and returns their results.

    Exceptions of type ``base_exception`` are caught and returned
    in the results list. Traces are printed to stderr.

    This is a bit faster and nicer to use than ``asyncio.gather``.
    In the case there is only a single awaitable, it is awaited
    directly rather than creating and awaiting a new task.
    """

    if not isinstance(awaitables, Sized):
        awaitables = list(awaitables)

    if len(awaitables) == 1:
        try:
            return [await awaitables[0]]
        except base_exception as e:
            traceback.print_exc()
            return [e]

    tasks = [
        a if isinstance(a, Task) else asyncio.create_task(a)
        for a in awaitables
    ]

    results = []
    for task in tasks:
        try:
            results.append(await task)
        except base_exception as e:
            traceback.print_exc()
            results.append(e)

    return results


def noop():
    """Does nothing. Use to return control to the event loop."""
    return asyncio.sleep(0)


class Reader(Protocol):
    async def readexactly(self, n: int) -> bytes:
        ...

    async def readuntil(self, separator: bytes) -> bytes:
        ...


class Writer(Protocol):
    def write(self, data: bytes) -> None:
        ...

    async def drain(self) -> None:
        ...

    def close(self) -> None:
        ...

    def is_closing(self) -> bool:
        ...

    async def wait_closed(self) -> None:
        ...

    def get_extra_info(self, name: str, default=None):
        ...


class LockableWriter(Writer):
    def __init__(self, writer: Writer):
        self.writer = writer
        self._lock = Lock()

    @property
    def lock(self) -> Lock:
        return self._lock

    async def __aenter__(self) -> 'LockableWriter':
        await self.lock.acquire()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        self._lock.release()

    def write(self, data: bytes) -> None:
        self.require_locked()
        self.writer.write(data)

    async def drain(self) -> None:
        await self.writer.drain()

    def close(self) -> None:
        self.writer.close()

    def is_closing(self) -> bool:
        return self.writer.is_closing()

    async def wait_closed(self) -> None:
        await self.writer.wait_closed()

    def get_extra_info(self, name: str, default=None):
        return self.writer.get_extra_info(name, default)

    def require_locked(self) -> None:
        if not self._lock.locked():
            raise RuntimeError('Writer must be locked before writing')


class BufferReader(Reader):
    def __init__(self, data: bytes):
        self._data = BytesIO(data)

    async def readexactly(self, n: int) -> bytes:
        data = self._data.read(n)
        assert len(data) == n
        return data

    async def readuntil(self, separator: bytes) -> bytes:
        raise NotImplementedError()


class BufferWriter(bytearray, Buffer, Writer):
    def write(self, data: bytes) -> None:
        self.extend(data)

    async def drain(self) -> None:
        pass

    def close(self) -> None:
        pass

    def is_closing(self) -> bool:
        pass

    async def wait_closed(self) -> None:
        pass

    def get_extra_info(self, name: str, default=None):
        raise NotImplementedError()



import logging
from argparse import ArgumentParser, Namespace
from datetime import datetime

from rosy import build_node_from_args
from rosy.argparse import add_node_args
from rosy.cli.topic.utils import print_args_and_kwargs
from rosy.cli.utils import add_log_arg
from rosy.types import Topic


async def echo_main(args: Namespace):
    logging.basicConfig(level=args.log)

    node = await build_node_from_args(args=args)

    for topic in args.topics:
        await node.listen(topic, handle_message)

    print(f'Listening to topics: {args.topics}')
    await node.forever()


async def handle_message(topic: Topic, *args, **kwargs):
    now = datetime.now()
    print(f'[{now}]')

    print(f'topic={topic!r}')
    print_args_and_kwargs(args, kwargs)
    print()


def add_echo_command(subparsers) -> None:
    parser: ArgumentParser = subparsers.add_parser(
        'echo',
        description='Start a node that listens to topics and prints received messages.',
        help='listen to topics',
    )

    parser.add_argument(
        'topics',
        nargs='+',
        metavar='topic',
        help='The topic(s) to listen to.',
    )

    add_log_arg(parser)

    add_node_args(
        parser,
        default_node_name='rosy topic echo',
    )



from argparse import ArgumentParser, Namespace

from rosy.cli.node.list import add_list_command, list_main


async def node_main(args: Namespace) -> None:
    if args.node_command == 'list':
        await list_main(args)
    else:
        raise ValueError(f'Unknown node command: {args.node_command}')


def add_node_command(subparsers) -> None:
    parser: ArgumentParser = subparsers.add_parser(
        'node',
        description='Node commands.',
        help='Node commands like list',
    )

    subparsers = parser.add_subparsers(
        title='commands',
        dest='node_command',
        required=True,
    )

    add_list_command(subparsers)



from collections import defaultdict

from rosy.specs import MeshNodeSpec, MeshTopologySpec
from rosy.types import Service, Topic


class MeshTopologyManager:
    def __init__(self):
        self._topology: MeshTopologySpec
        self._topic_nodes: dict[Topic, list[MeshNodeSpec]]
        self._service_nodes: dict[Service, list[MeshNodeSpec]]

        self.set_topology(MeshTopologySpec(nodes=[]))

    @property
    def topology(self) -> MeshTopologySpec:
        return self._topology

    def set_topology(self, value: MeshTopologySpec) -> None:
        self._topology = value

        self._cache_topic_nodes()
        self._cache_service_nodes()

    def _cache_topic_nodes(self) -> None:
        topic_nodes = defaultdict(list)

        for node in self.topology.nodes:
            for topic in node.topics:
                topic_nodes[topic].append(node)

        self._topic_nodes_cache = topic_nodes

    def _cache_service_nodes(self) -> None:
        service_nodes = defaultdict(list)

        for node in self.topology.nodes:
            for service in node.services:
                service_nodes[service].append(node)

        self._service_nodes_cache = service_nodes

    def get_nodes_listening_to_topic(self, topic: Topic) -> list[MeshNodeSpec]:
        return self._topic_nodes_cache[topic]

    def get_nodes_providing_service(self, service: str) -> list[MeshNodeSpec]:
        return self._service_nodes_cache[service]

    def get_removed_nodes(
            self,
            new_topology: MeshTopologySpec,
    ) -> list[MeshNodeSpec]:
        """
        Returns a list of nodes that were removed in the new topology.
        """

        new_node_ids = {node.id for node in new_topology.nodes}

        return [
            node for node in self.topology.nodes
            if node.id not in new_node_ids
        ]



from collections.abc import Collection
from dataclasses import dataclass, field
from socket import AddressFamily
from uuid import UUID, uuid1

from rosy.network import get_hostname
from rosy.types import Host, Port, Service, Topic


@dataclass
class IpConnectionSpec:
    host: Host
    port: Port
    family: AddressFamily


@dataclass
class UnixConnectionSpec:
    path: str
    host: Host = field(default_factory=get_hostname)


ConnectionSpec = IpConnectionSpec | UnixConnectionSpec

NodeName = str
NodeUUID = UUID


@dataclass(order=True, frozen=True)
class NodeId:
    name: NodeName
    hostname: Host = field(default_factory=get_hostname)
    uuid: NodeUUID = field(default_factory=uuid1)

    def __str__(self) -> str:
        name = repr(self.name) if ' ' in self.name else self.name
        short_uuid = str(self.uuid)[:4]
        return f'{name}@{self.hostname} ({short_uuid})'


@dataclass
class MeshNodeSpec:
    id: NodeId
    connection_specs: list[ConnectionSpec]
    topics: set[Topic]
    services: set[Service]


@dataclass
class MeshTopologySpec:
    nodes: Collection[MeshNodeSpec]



from dataclasses import dataclass

from rosy.specs import MeshNodeSpec, MeshTopologySpec


@dataclass
class RegisterNodeRequest:
    node_spec: MeshNodeSpec


class RegisterNodeResponse:
    pass


@dataclass
class MeshTopologyBroadcast:
    mesh_topology: MeshTopologySpec



import time
from abc import ABC, abstractmethod
from collections import defaultdict
from collections.abc import Callable
from itertools import chain, groupby
from random import Random
from typing import Any

from rosy.specs import MeshNodeSpec, NodeId
from rosy.types import Service, Topic

GroupKey = Callable[[MeshNodeSpec], Any]


class TopicLoadBalancer(ABC):
    @abstractmethod
    def choose_nodes(self, nodes: list[MeshNodeSpec], topic: Topic) -> list[MeshNodeSpec]:
        """
        Takes a list of nodes listening to the given topic
        and returns which nodes to send the message to.
        """
        ...  # pragma: no cover


class ServiceLoadBalancer(ABC):
    @abstractmethod
    def choose_node(self, nodes: list[MeshNodeSpec], service: Service) -> MeshNodeSpec | None:
        """
        Takes a list of nodes providing the given service
        and returns which node to send the request to.
        """
        ...  # pragma: no cover


class NoopTopicLoadBalancer(TopicLoadBalancer):
    """No load balancing. Sends to all nodes."""

    def choose_nodes(self, nodes: list[MeshNodeSpec], topic: Topic) -> list[MeshNodeSpec]:
        return nodes


def node_name_group_key(node: MeshNodeSpec) -> str:
    return node.id.name


class GroupingTopicLoadBalancer(TopicLoadBalancer):
    """
    Groups nodes according to ``group_key`` and applies
    the given load balancer to each group.
    """

    def __init__(
            self,
            group_key: GroupKey,
            load_balancer: TopicLoadBalancer,
    ):
        self.group_key = group_key
        self.load_balancer = load_balancer

    def choose_nodes(self, nodes: list[MeshNodeSpec], topic: Topic) -> list[MeshNodeSpec]:
        if not nodes:
            return []

        nodes = sorted(nodes, key=self.group_key)
        grouped_nodes = (list(group) for _, group in groupby(nodes, key=self.group_key))
        return list(chain.from_iterable(
            self.load_balancer.choose_nodes(group, topic)
            for group in grouped_nodes
        ))


class RandomLoadBalancer(TopicLoadBalancer, ServiceLoadBalancer):
    """Chooses a single node at random."""

    def __init__(self, rng: Random = None):
        self.rng = rng or Random()

    def choose_nodes(self, nodes: list[MeshNodeSpec], topic: Topic) -> list[MeshNodeSpec]:
        return [self.rng.choice(nodes)] if nodes else []

    def choose_node(self, nodes: list[MeshNodeSpec], service: Service) -> MeshNodeSpec | None:
        return self.rng.choice(nodes) if nodes else None


class LeastRecentLoadBalancer(TopicLoadBalancer, ServiceLoadBalancer):
    """Chooses the node that has been least recently chosen."""

    def __init__(
            self,
            time_func: Callable[[], float | int] = time.monotonic_ns,
            rng: Random = None,
    ):
        self.time_func = time_func
        self.rng = rng or Random()

        # By defaulting the last used time to a random value before the current time,
        # we ensure that any ties between new nodes will be broken randomly,
        # while still ensuring new nodes will be chosen first.
        t0 = time_func()
        self._last_used: defaultdict[NodeId, float | int] = defaultdict(
            lambda: t0 * self.rng.random(),
        )

    def choose_nodes(self, nodes: list[MeshNodeSpec], topic: Topic) -> list[MeshNodeSpec]:
        return [self._get_least_recent_node(nodes)] if nodes else []

    def choose_node(self, nodes: list[MeshNodeSpec], service: Service) -> MeshNodeSpec | None:
        return self._get_least_recent_node(nodes) if nodes else None

    def _get_least_recent_node(self, nodes: list[MeshNodeSpec]) -> MeshNodeSpec:
        last_used_times = (self._last_used[node.id] for node in nodes)

        node, _ = min(
            zip(nodes, last_used_times),
            key=lambda i: i[1],
        )

        self._last_used[node.id] = self.time_func()
        return node



import logging
from argparse import ArgumentParser, Namespace

from rosy.argparse import add_authkey_arg, add_coordinator_arg
from rosy.cli.utils import add_log_arg, get_mesh_topology
from rosy.specs import MeshNodeSpec


async def list_main(args: Namespace):
    logging.basicConfig(level=args.log)

    topology = await get_mesh_topology(args)

    nodes = sorted(topology.nodes, key=lambda n: n.id)

    print(f'{len(nodes)} nodes:')
    for node in nodes:
        print()
        print_node(node, args.verbose)


def print_node(node: MeshNodeSpec, verbose: bool) -> None:
    print(node.id)

    if verbose:
        print(f'- UUID: {node.id.uuid}')

    if node.topics:
        print('- topics:')
        for topic in sorted(node.topics):
            print(f'  - {topic!r}')
    else:
        print('- topics: none')

    if node.services:
        print('- services:')
        for service in sorted(node.services):
            print(f'  - {service!r}')
    else:
        print('- services: none')

    if verbose:
        print('- supported connection methods:')
        for conn in node.connection_specs:
            print(f'  - {conn}')


def add_list_command(subparsers) -> None:
    parser: ArgumentParser = subparsers.add_parser(
        'list',
        description='List all nodes in the mesh.',
        help='list nodes in the mesh',
    )

    parser.add_argument(
        '--verbose', '-v',
        action='store_true',
        help='Print all info about each node, including UUIDs and connection methods.',
    )

    add_log_arg(parser)
    add_coordinator_arg(parser)
    add_authkey_arg(parser)



from argparse import Namespace
from typing import Literal

from rosy import Node
from rosy.argparse import get_node_arg_parser
from rosy.authentication import Authenticator, optional_authkey_authenticator
from rosy.codec import (
    Codec,
    DictCodec,
    FixedLengthIntCodec,
    LengthPrefixedStringCodec,
    SequenceCodec,
    json_codec,
    msgpack_codec,
    pickle_codec,
)
from rosy.coordinator.client import build_coordinator_client
from rosy.coordinator.constants import DEFAULT_COORDINATOR_PORT
from rosy.network import get_lan_hostname
from rosy.node.clienthandler import ClientHandler
from rosy.node.codec import NodeMessageCodec
from rosy.node.loadbalancing import (
    GroupingTopicLoadBalancer,
    LeastRecentLoadBalancer,
    ServiceLoadBalancer,
    TopicLoadBalancer,
    node_name_group_key,
)
from rosy.node.peer import PeerConnectionBuilder, PeerConnectionManager, PeerSelector
from rosy.node.servers import TcpServerProvider, ServerProvider, ServersManager, TmpUnixServerProvider
from rosy.node.service.caller import ServiceCaller
from rosy.node.service.codec import ServiceRequestCodec, ServiceResponseCodec
from rosy.node.service.handlermanager import ServiceHandlerManager
from rosy.node.service.requesthandler import ServiceRequestHandler
from rosy.node.topic.codec import TopicMessageCodec
from rosy.node.topic.listenermanager import TopicListenerManager
from rosy.node.topic.messagehandler import TopicMessageHandler
from rosy.node.topic.sender import TopicSender
from rosy.node.topology import MeshTopologyManager
from rosy.specs import NodeId
from rosy.types import Data, Host, Port, ServerHost


async def build_node_from_args(
        default_node_name: str = None,
        args: Namespace = None,
        **kwargs,
) -> Node:
    """
    Builds a node from command line arguments.

    Args:
        default_node_name:
            Default node name. If not given, the argument is required.
            Ignored if `args` is given.
        args:
            Arguments from an argument parser. If not given, an argument parser
            is created using `get_node_arg_parser` and is used to parse args.
            This is useful if you create your own argument parser.
        kwargs:
            Additional keyword arguments to pass to `build_node`.
            These will override anything specified in `args`.
    """

    if args is None:
        args = get_node_arg_parser(default_node_name).parse_args()

    build_args = vars(args)

    if hasattr(args, 'coordinator'):
        build_args['coordinator_host'] = args.coordinator.host
        build_args['coordinator_port'] = args.coordinator.port

    build_args.update(kwargs)

    return await build_node(**build_args)


async def build_node(
        name: str,
        coordinator_host: Host = 'localhost',
        coordinator_port: Port = DEFAULT_COORDINATOR_PORT,
        coordinator_reconnect_timeout: float | None = 5.0,
        allow_unix_connections: bool = True,
        allow_tcp_connections: bool = True,
        node_server_host: ServerHost = None,
        node_client_host: Host = None,
        data_codec: Codec[Data] | Literal['pickle', 'json', 'msgpack'] = 'pickle',
        authkey: bytes = None,
        authenticator: Authenticator = None,
        topic_load_balancer: TopicLoadBalancer = None,
        service_load_balancer: ServiceLoadBalancer = None,
        start: bool = True,
        **kwargs,
) -> Node:
    authenticator = authenticator or optional_authkey_authenticator(authkey)

    coordinator_client = await build_coordinator_client(
        coordinator_host,
        coordinator_port,
        authenticator,
        reconnect_timeout=coordinator_reconnect_timeout,
    )

    topology_manager = MeshTopologyManager()

    peer_selector = build_peer_selector(
        topology_manager,
        topic_load_balancer,
        service_load_balancer,
    )

    connection_manager = PeerConnectionManager(
        PeerConnectionBuilder(authenticator),
    )

    if data_codec == 'pickle':
        data_codec = pickle_codec
    elif data_codec == 'json':
        data_codec = json_codec
    elif data_codec == 'msgpack':
        data_codec = msgpack_codec

    node_message_codec = build_node_message_codec(data_codec)

    topic_sender = TopicSender(peer_selector, connection_manager, node_message_codec)

    topic_listener_manager = TopicListenerManager()
    topic_message_handler = TopicMessageHandler(topic_listener_manager)

    service_handler_manager = ServiceHandlerManager()
    service_request_handler = ServiceRequestHandler(
        service_handler_manager,
        node_message_codec,
    )

    client_handler = ClientHandler(
        authenticator,
        node_message_codec,
        topic_message_handler,
        service_request_handler,
    )

    server_providers = build_server_providers(
        allow_unix_connections,
        allow_tcp_connections,
        node_server_host,
        node_client_host,
    )
    servers_manager = ServersManager(server_providers, client_handler.handle_client)

    request_id_bytes = 2  # Codec uses 2 bytes for request ID
    service_caller = ServiceCaller(
        peer_selector,
        connection_manager,
        node_message_codec,
        max_request_ids=2 ** (8 * request_id_bytes),
    )

    node = Node(
        id=NodeId(name),
        coordinator_client=coordinator_client,
        servers_manager=servers_manager,
        topology_manager=topology_manager,
        connection_manager=connection_manager,
        topic_sender=topic_sender,
        topic_listener_manager=topic_listener_manager,
        service_caller=service_caller,
        service_handler_manager=service_handler_manager,
    )

    if start:
        await node.start()

    return node


def build_server_providers(
        allow_unix_connections: bool,
        allow_tcp_connections: bool,
        node_server_host: ServerHost | None,
        node_client_host: Host | None,
) -> list[ServerProvider]:
    server_providers = []

    if allow_unix_connections:
        server_providers.append(TmpUnixServerProvider())

    if allow_tcp_connections:
        if not node_client_host:
            node_client_host = get_lan_hostname()

        provider = TcpServerProvider(node_server_host, node_client_host)
        server_providers.append(provider)

    if not server_providers:
        raise ValueError('Must allow at least one type of connection')

    return server_providers


def build_peer_selector(
        topology_manager: MeshTopologyManager,
        topic_load_balancer: TopicLoadBalancer | None,
        service_load_balancer: ServiceLoadBalancer | None,
) -> PeerSelector:
    least_recent_load_balancer = LeastRecentLoadBalancer()

    default_topic_load_balancer = GroupingTopicLoadBalancer(
        group_key=node_name_group_key,
        load_balancer=least_recent_load_balancer,
    )

    return PeerSelector(
        topology_manager,
        topic_load_balancer=topic_load_balancer or default_topic_load_balancer,
        service_load_balancer=service_load_balancer or least_recent_load_balancer,
    )


def build_node_message_codec(
        data_codec: Codec[Data],
) -> NodeMessageCodec:
    short_string_codec = LengthPrefixedStringCodec(
        len_prefix_codec=FixedLengthIntCodec(length=1)
    )

    short_int_codec = FixedLengthIntCodec(length=1)

    args_codec: SequenceCodec[Data] = SequenceCodec(
        len_header_codec=short_int_codec,
        item_codec=data_codec,
    )

    kwargs_codec: DictCodec[str, Data] = DictCodec(
        len_header_codec=short_int_codec,
        key_codec=short_string_codec,
        value_codec=data_codec,
    )

    request_id_codec = FixedLengthIntCodec(length=2)

    return NodeMessageCodec(
        topic_message_codec=TopicMessageCodec(
            topic_codec=short_string_codec,
            args_codec=args_codec,
            kwargs_codec=kwargs_codec,
        ),
        service_request_codec=ServiceRequestCodec(
            request_id_codec,
            service_codec=short_string_codec,
            args_codec=args_codec,
            kwargs_codec=kwargs_codec,
        ),
        service_response_codec=ServiceResponseCodec(
            request_id_codec,
            data_codec=data_codec,
            error_codec=LengthPrefixedStringCodec(
                len_prefix_codec=FixedLengthIntCodec(length=2),
            )
        ),
    )



from pathlib import Path

import yaml

from rosy.network import get_hostname


def load_config(path: Path) -> dict:
    with path.open('r') as file:
        return yaml.safe_load(file)


def is_enabled(config: dict) -> bool:
    disabled = config.get('disabled', False)
    return not disabled and is_enabled_on_host(config)


def is_enabled_on_host(config: dict) -> bool:
    hostname = get_hostname()
    on_host = config.get('on_host', hostname)
    return on_host == hostname



import logging
from argparse import ArgumentParser, Namespace

from rosy.argparse import add_authkey_arg, add_coordinator_arg
from rosy.authentication import optional_authkey_authenticator
from rosy.cli.utils import add_log_arg, get_mesh_topology
from rosy.coordinator.client import build_coordinator_client


async def list_main(args: Namespace):
    logging.basicConfig(level=args.log)

    topology = await get_mesh_topology(args)

    services = sorted({
        service
        for node in topology.nodes
        for service in node.services
    })

    for service in services:
        print(service)


def add_list_command(subparsers) -> None:
    parser: ArgumentParser = subparsers.add_parser(
        'list',
        description='List all services currently being provided by nodes.',
        help='list services being provided',
    )

    add_log_arg(parser)

    add_coordinator_arg(parser)
    add_authkey_arg(parser)



import socket

from rosy.types import Host


def get_hostname() -> Host:
    """Return the current host name."""
    return socket.gethostname()


def get_lan_hostname(suffix: str = '.local') -> Host:
    """
    Return the mDNS hostname of this machine as seen on the local network,
    e.g. "<hostname>.local".
    """
    return get_hostname() + suffix



import argparse
import asyncio
import importlib
import logging
import re
from argparse import ArgumentParser, Namespace
from ast import literal_eval
from collections.abc import Iterable
from datetime import datetime
from typing import Any

from rosy import build_node_from_args
from rosy.argparse import add_node_args
from rosy.cli.topic.utils import print_args_and_kwargs
from rosy.cli.utils import add_log_arg


async def send_main(args: Namespace):
    logging.basicConfig(level=args.log)

    node = await build_node_from_args(args=args)

    topic = node.get_topic(args.topic)

    async def send_once():
        if not args.no_wait and not await topic.has_listeners():
            print(f'Waiting for listeners...')
            await topic.wait_for_listener()

        topic_args, topic_kwargs = parse_args_and_kwargs(args.args)

        now = datetime.now()
        print(f'[{now}]')
        print(f'Sending to topic={args.topic!r}')
        print_args_and_kwargs(topic_args, topic_kwargs)
        print()

        await topic.send(*topic_args, **topic_kwargs)

    if args.interval < 0:
        await send_once()
        return

    while True:
        await send_once()
        await asyncio.sleep(args.interval)


def parse_args_and_kwargs(args: Iterable[str]) -> tuple[list[Any], dict[str, Any]]:
    topic_args = []
    topic_kwargs = {}

    for arg in args:
        key, value = key_value_from_str(arg)

        if key is None:
            if topic_kwargs:
                raise ValueError(
                    f'Positional argument {arg!r} must come '
                    f'before the keyword arguments.'
                )

            topic_args.append(value)
        else:
            topic_kwargs[key] = value

    return topic_args, topic_kwargs


def key_value_from_str(value: str) -> tuple[str | None, Any]:
    match = re.match(
        r'^([A-Za-z_][\w_]*)=(.*)$',
        value,
    )

    if match:
        key, value = match.groups()
        return key, value_from_str(value)
    else:
        return None, value_from_str(value)


def value_from_str(value: str) -> Any:
    match = re.match(
        # prefix
        r'^call:'
        # module name
        r'([\w.]+)\.'
        # callable name
        r'(\w+)'
        # args
        r'\((.*)\)'
        r'$',
        value
    )

    if not match:
        return literal_eval(value)

    module, callable_, args = match.groups()
    module = importlib.import_module(module)
    callable_ = getattr(module, callable_)
    result = eval(f'callable({args})', {}, dict(callable=callable_))
    return result


def add_send_command(subparsers) -> None:
    parser: ArgumentParser = subparsers.add_parser(
        'send',
        description='''
Start a node that sends messages to a topic.

---

Simple example:
    rosy topic send my-topic "'hello world'" 42 pi=3.14

This will send a message to the topic `my-topic` with args `'hello world'` and 42, and keyword arg `pi=3.14`.

---

Intermediate example:
    rosy topic send my-topic "{'key': 'value', 'data': [1, 2, 3]}"

This will send a message with a dictionary as an argument.

---

Advanced example:
    rosy topic send my-topic "call:myproj.MyData('data', pi=3.14)" "send_time=call:time.time()"

This will call the specified class/function to populate the arguments.

---
'''.strip(),
        help='send messages to a topic',
        formatter_class=argparse.RawTextHelpFormatter,
    )

    parser.add_argument(
        'topic',
        help='The topic to send to.',
    )

    parser.add_argument(
        'args',
        nargs='*',
        metavar='arg/kwarg',
        help='''
Arg(s) and/or kwarg(s) to send to the topic.
Each arg must be a valid Python expression,
or follow the format:
`call:module.callable(*args, **kwargs)`
'''.strip(),
    )

    parser.add_argument(
        '--interval', '-i',
        default=-1,
        type=float,
        help='The interval in seconds to send messages. A value < 0 will '
             'cause the message to be sent only once. Default: %(default)s',
    )

    parser.add_argument(
        '--no-wait',
        action='store_true',
        help='Send the message without waiting for any listeners',
    )

    add_log_arg(parser)

    add_node_args(
        parser,
        default_node_name='rosy topic send',
    )



from argparse import ArgumentParser, Namespace

from rosy.authentication import optional_authkey_authenticator
from rosy.coordinator.client import build_coordinator_client
from rosy.specs import MeshTopologySpec


def add_log_arg(parser: ArgumentParser, default: str = 'WARNING') -> None:
    parser.add_argument(
        '--log',
        default=default,
        help='Log level; DEBUG, INFO, ERROR, etc. Default: %(default)s'
    )


async def get_mesh_topology(args: Namespace) -> MeshTopologySpec:
    authenticator = optional_authkey_authenticator(args.authkey)

    coordinator_client = await build_coordinator_client(
        host=args.coordinator.host,
        port=args.coordinator.port,
        authenticator=authenticator,
        reconnect_timeout=None,
    )

    return await coordinator_client.get_topology()



import asyncio
import logging
from abc import ABC, abstractmethod
from asyncio import StreamReader
from codecs import StreamWriter

from rosy.asyncio import close_ignoring_errors, many
from rosy.authentication import AuthKey, Authenticator, optional_authkey_authenticator
from rosy.coordinator.constants import DEFAULT_COORDINATOR_HOST, DEFAULT_COORDINATOR_PORT
from rosy.objectio import CodecObjectReader, CodecObjectWriter, ObjectIO
from rosy.reqres import MeshTopologyBroadcast, RegisterNodeRequest, RegisterNodeResponse
from rosy.rpc import ObjectIORPC, RPC
from rosy.specs import MeshNodeSpec, MeshTopologySpec, NodeId
from rosy.types import Port, ServerHost

logger = logging.getLogger(__name__)


class MeshCoordinatorServer(ABC):
    @abstractmethod
    async def start(self) -> None:
        ...


class RPCMeshCoordinatorServer(MeshCoordinatorServer):
    def __init__(
            self,
            start_stream_server,
            build_rpc,
            authenticator: Authenticator,
    ):
        self.start_stream_server = start_stream_server
        self.build_rpc = build_rpc
        self.authenticator = authenticator

        self._node_clients: dict[RPC, NodeId | None] = {}
        self._nodes: dict[NodeId, MeshNodeSpec] = {}

    async def start(self) -> None:
        server = await self.start_stream_server(self._handle_connection)

    async def _handle_connection(self, reader: StreamReader, writer: StreamWriter) -> None:
        peer_name = writer.get_extra_info('peername') or writer.get_extra_info('sockname')
        logger.debug(f'New connection from: {peer_name}')

        await self.authenticator.authenticate(reader, writer)

        rpc = self.build_rpc(reader, writer)
        rpc.request_handler = lambda r: self._handle_request(r, rpc, peer_name)
        self._node_clients[rpc] = None

        try:
            await rpc.run_forever()
        except (ConnectionResetError, EOFError):
            logger.debug(f'Client disconnected: {peer_name}')
        finally:
            try:
                await self._remove_node(rpc)
            finally:
                await close_ignoring_errors(writer)

    async def _handle_request(self, request, rpc: RPC, peer_name: str):
        if request == b'get_topology':
            return self._get_mesh_topology()
        elif request == b'ping':
            logger.debug(f'Received heartbeat from: {peer_name}')
            return b'pong'
        elif isinstance(request, RegisterNodeRequest):
            return await self._handle_register_node(request, rpc)
        else:
            raise Exception(f'Received invalid request object of type={type(request)}')

    def _get_mesh_topology(self) -> MeshTopologySpec:
        nodes = sorted(self._nodes.values(), key=lambda n: n.id)
        return MeshTopologySpec(nodes=nodes)

    async def _handle_register_node(
            self,
            request: RegisterNodeRequest,
            rpc: RPC,
    ) -> RegisterNodeResponse:
        logger.debug(f'Got register node request: {request}')

        node_spec = request.node_spec
        node_is_new = node_spec.id not in self._nodes
        self._node_clients[rpc] = node_spec.id
        self._nodes[node_spec.id] = node_spec

        if node_is_new:
            logger.info(f'Node registered: {node_spec.id}')
            logger.info(f'Total nodes: {len(self._nodes)}')
        else:
            logger.debug(f'Node re-registered: {node_spec.id}')

        await self._broadcast_topology()

        return RegisterNodeResponse()

    async def _remove_node(self, rpc: RPC) -> None:
        node_id = self._node_clients.pop(rpc)
        if node_id is None:
            return

        self._nodes.pop(node_id, None)
        logger.info(f'Node removed: {node_id}')
        logger.info(f'Total nodes: {len(self._nodes)}')

        await self._broadcast_topology()

    async def _broadcast_topology(self) -> None:
        logger.debug(f'Broadcasting topology to {len(self._nodes)} nodes...')
        if not self._nodes:
            return

        mesh_topology = self._get_mesh_topology()
        for node in mesh_topology.nodes:
            logger.debug(f'- {node.id}: {node}')

        message = MeshTopologyBroadcast(mesh_topology)

        await many(
            node_client.send_message(message)
            for node_client in self._node_clients.keys()
        )


def build_mesh_coordinator_server(
        host: ServerHost = DEFAULT_COORDINATOR_HOST,
        port: Port = DEFAULT_COORDINATOR_PORT,
        authkey: AuthKey = None,
        authenticator: Authenticator = None,
) -> MeshCoordinatorServer:
    async def start_stream_server(cb):
        return await asyncio.start_server(cb, host=host, port=port)

    def build_rpc(reader, writer):
        return ObjectIORPC(ObjectIO(
            reader=CodecObjectReader(reader),
            writer=CodecObjectWriter(writer),
        ))

    authenticator = authenticator or optional_authkey_authenticator(authkey)

    return RPCMeshCoordinatorServer(
        start_stream_server,
        build_rpc,
        authenticator,
    )



import asyncio
import logging
from functools import wraps
from typing import NamedTuple

from rosy.asyncio import forever
from rosy.coordinator.client import MeshCoordinatorClient
from rosy.node.peer import PeerConnectionManager
from rosy.node.servers import ServersManager
from rosy.node.service.caller import ServiceCaller
from rosy.node.service.handlermanager import ServiceHandlerManager
from rosy.node.topic.listenermanager import TopicListenerManager
from rosy.node.topic.sender import TopicSender
from rosy.node.topology import MeshTopologyManager
from rosy.reqres import MeshTopologyBroadcast
from rosy.specs import MeshNodeSpec, NodeId
from rosy.types import Data, Service, ServiceCallback, Topic, TopicCallback

logger = logging.getLogger(__name__)


class Node:
    def __init__(
            self,
            id: NodeId,
            coordinator_client: MeshCoordinatorClient,
            servers_manager: ServersManager,
            topology_manager: MeshTopologyManager,
            connection_manager: PeerConnectionManager,
            topic_sender: TopicSender,
            topic_listener_manager: TopicListenerManager,
            service_caller: ServiceCaller,
            service_handler_manager: ServiceHandlerManager,
    ):
        """
        This is a node on the mesh. It is responsible for sending and receiving
        messages on topics and services.

        You should not instantiate this class directly; instead, use
        `rosy.build_node()` or `rosy.build_node_from_args()`.
        """

        self._id = id
        self.coordinator_client = coordinator_client
        self.servers_manager = servers_manager
        self.topology_manager = topology_manager
        self.connection_manager = connection_manager
        self.topic_sender = topic_sender
        self.topic_listener_manager = topic_listener_manager
        self.service_caller = service_caller
        self.service_handler_manager = service_handler_manager

        self._started: bool = False

        coordinator_client.set_broadcast_handler(self._handle_topology_broadcast)

    @property
    def id(self) -> NodeId:
        return self._id

    def __str__(self) -> str:
        return str(self.id)

    async def start(self) -> None:
        """
        Start the node by starting the servers and registering with the coordinator.
        """

        if self._started:
            raise RuntimeError('Node is already started.')

        logger.info(f'Starting node {self.id}')

        logger.debug('Starting servers')
        await self.servers_manager.start_servers()

        await self.register()

        self._started = True

    async def send(self, topic: Topic, *args: Data, **kwargs: Data) -> None:
        """Send a message on a topic, with optional arguments and keyword arguments."""
        await self.topic_sender.send(topic, args, kwargs)

    async def listen(
            self,
            topic: Topic,
            callback: TopicCallback,
    ) -> None:
        """Start listening to a topic with a callback function."""
        self.topic_listener_manager.set_callback(topic, callback)
        await self.register()

    async def stop_listening(self, topic: Topic) -> None:
        """Stop listening to a topic."""

        callback = self.topic_listener_manager.remove_callback(topic)

        if callback is not None:
            await self.register()
        else:
            logger.warning(f"Attempted to remove non-existing listener for topic={topic!r}")

    async def topic_has_listeners(self, topic: Topic) -> bool:
        """Check if there are any listeners for a topic."""
        listeners = self.topology_manager.get_nodes_listening_to_topic(topic)
        return bool(listeners)

    async def wait_for_listener(self, topic: Topic, poll_interval: float = 1.) -> None:
        """
        Wait until there is a listener for a topic.

        Useful for send-only nodes to avoid doing unnecessary work when there
        are no listeners for a topic.

        Combine this with ``depends_on_listener`` in intermediate nodes to make all
        nodes in a chain wait until there is a listener at the end of the chain.
        """

        while not await self.topic_has_listeners(topic):
            await asyncio.sleep(poll_interval)

    def depends_on_listener(self, downstream_topic: Topic, poll_interval: float = 1.):
        """
        Decorator for callback functions that send messages to a downstream
        topic. If there is no listener for the downstream topic, then the node
        will stop listening to the upstream topic until there is a listener for
        the downstream topic.

        Useful for nodes that do intermediate processing, i.e. nodes that
        receive a message on a topic, process it, and then send the result on
        another topic.

        Example:
            >>> @node.depends_on_listener('bar')
            >>> async def handle_foo(topic, data):
            >>>     await node.send('bar', data)
            >>>
            >>> await node.listen('foo', handle_foo)

        Combine this with ``wait_for_listener`` in send-only nodes to make all
        nodes in a chain wait until there is a listener at the end of the chain.
        """

        def decorator(callback):
            @wraps(callback)
            async def wrapper(topic: Topic, data: Data) -> None:
                if await self.topic_has_listeners(downstream_topic):
                    await callback(topic, data)
                    return

                await self.stop_listening(topic)

                async def wait_for_listener_then_listen():
                    await self.wait_for_listener(downstream_topic, poll_interval)
                    await self.listen(topic, wrapper)

                asyncio.create_task(wait_for_listener_then_listen())

            return wrapper

        return decorator

    def get_topic(self, topic: Topic) -> 'TopicProxy':
        """
        Returns a topic proxy that can be used to interact with a topic (e.g.
        send messages) without needing to pass the topic name each time.

        Example:
            >>> topic = node.get_topic('my_topic')
            >>> await topic.send('Hello, world!')
            >>> # ... is equivalent to ...
            >>> await node.send('my_topic', 'Hello, world!')
        """
        return TopicProxy(self, topic)

    async def call(self, service: Service, *args, **kwargs) -> Data:
        """Call a service and return the result."""
        return await self.service_caller.call(service, args, kwargs)

    async def add_service(self, service: Service, handler: ServiceCallback) -> None:
        """Add a service to the node that other nodes can call."""
        self.service_handler_manager.set_callback(service, handler)
        await self.register()

    async def remove_service(self, service: Service) -> None:
        """Stop providing a service."""
        callback = self.service_handler_manager.remove_callback(service)

        if callback is not None:
            await self.register()
        else:
            logger.warning(f"Attempted to remove non-existing service={service!r}")

    async def service_has_providers(self, service: Service) -> bool:
        """Check if there are any nodes that provide the service."""
        providers = self.topology_manager.get_nodes_providing_service(service)
        return bool(providers)

    async def wait_for_service_provider(self, service: Service, poll_interval: float = 1.) -> None:
        """Wait until there is a provider for a service."""
        while not await self.service_has_providers(service):
            await asyncio.sleep(poll_interval)

    def get_service(self, service: Service) -> 'ServiceProxy':
        """
        Returns a convenient way to call a service if used more than once.

        Example:
            >>> math_service = node.get_service('math')
            >>> result = await math_service('2 + 2')
            >>> # ... is equivalent to ...
            >>> result = await node.call('math', '2 + 2')
        """

        return ServiceProxy(self, service)

    async def register(self) -> None:
        """
        Register the node with the coordinator.

        This is done automatically when the node is started,
        and when topics or services are added or removed,
        so there should be no need to call this manually.
        """

        node_spec = self._build_node_spec()
        logger.info('Registering node with coordinator')
        logger.debug(f'node_spec={node_spec}')
        await self.coordinator_client.register_node(node_spec)

    def _build_node_spec(self) -> MeshNodeSpec:
        return MeshNodeSpec(
            id=self.id,
            connection_specs=self.servers_manager.connection_specs,
            topics=self.topic_listener_manager.keys,
            services=self.service_handler_manager.keys,
        )

    async def _handle_topology_broadcast(self, broadcast: MeshTopologyBroadcast) -> None:
        new_topology = broadcast.mesh_topology

        logger.debug(
            f'Received mesh topology broadcast with '
            f'{len(new_topology.nodes)} nodes.'
        )

        removed_nodes = self.topology_manager.get_removed_nodes(new_topology)
        logger.debug(
            f'Removed {len(removed_nodes)} nodes: '
            f'{[str(node.id) for node in removed_nodes]}'
        )

        self.topology_manager.set_topology(new_topology)

        for node in removed_nodes:
            await self.connection_manager.close_connection(node)

    async def forever(self) -> None:
        """
        Does nothing forever. Convenience method to prevent your main function
        from exiting while the node is running.
        """
        await forever()  # pragma: no cover


class TopicProxy(NamedTuple):
    node: Node
    topic: Topic

    def __str__(self):
        name = self.__class__.__name__
        return f'{name}(topic={self.topic!r})'

    async def send(self, *args: Data, **kwargs: Data) -> None:
        await self.node.send(self.topic, *args, **kwargs)

    async def has_listeners(self) -> bool:
        return await self.node.topic_has_listeners(self.topic)

    async def wait_for_listener(self, poll_interval: float = 1.) -> None:
        await self.node.wait_for_listener(self.topic, poll_interval)

    def depends_on_listener(self, poll_interval: float = 1.):
        return self.node.depends_on_listener(self.topic, poll_interval)


class ServiceProxy(NamedTuple):
    node: Node
    service: Service

    def __str__(self) -> str:
        name = self.__class__.__name__
        return f'{name}(service={self.service!r})'

    async def __call__(self, *args: Data, **kwargs: Data) -> Data:
        return await self.call(*args, **kwargs)

    async def call(self, *args: Data, **kwargs: Data) -> Data:
        return await self.node.call(self.service, *args, **kwargs)

    async def has_providers(self) -> bool:
        return await self.node.service_has_providers(self.service)

    async def wait_for_provider(self, poll_interval: float = 1.) -> None:
        await self.node.wait_for_service_provider(self.service, poll_interval)



import asyncio
import time
from argparse import ArgumentParser, Namespace

from rosy import Node, build_node
from rosy.argparse import add_authkey_arg, add_coordinator_arg
from rosy.types import Topic


async def speedtest_main(args: Namespace) -> None:
    load_balancer = 'default' if args.enable_load_balancer else None

    node = await build_node(
        name=f'speed-test/{args.role}',
        coordinator_host=args.coordinator.host,
        coordinator_port=args.coordinator.port,
        allow_unix_connections=not args.disable_unix,
        allow_tcp_connections=not args.disable_tcp,
        data_codec=args.codec,
        authkey=args.authkey,
        topic_load_balancer=load_balancer,
        service_load_balancer=load_balancer,
    )

    speed_tester = SpeedTest(node)

    topic = args.topic

    if args.role == 'recv':
        await speed_tester.receive(topic)
    elif args.role == 'send':
        print('Waiting for listeners...')
        await node.wait_for_listener(topic)

        print(f'Running speed test for {args.seconds}s '
              f'with message_size={args.message_size}...')
        mps = await speed_tester.measure_mps(
            topic,
            message_size=args.message_size,
            duration=args.seconds,
        )
        print(f'[{node}] mps={round(mps)}')
    else:
        raise ValueError(f'Invalid role={args.role}')


class SpeedTest:
    def __init__(self, node: Node):
        self.node = node

    async def measure_mps(
            self,
            topic: Topic,
            message_size: int = 0,
            duration: float = 10.,
            warmup: float | None = 1.,
    ) -> float:
        """Measure messages per second."""

        topic_sender = self.node.get_topic(topic)

        if not await topic_sender.has_listeners():
            raise ValueError(f'No listeners for topic={topic}')

        dummy_data = 'A' * message_size

        if warmup is not None and warmup > 0.:
            end_time = time.monotonic() + warmup
            while time.monotonic() < end_time:
                await self.node.send('warmup')

        message_count = 0
        start_time = time.monotonic()

        while (end_time := time.monotonic()) - start_time < duration:
            send_time = time.time()
            data = send_time, dummy_data
            await topic_sender.send(data)
            message_count += 1

        true_duration = end_time - start_time

        await self.node.send('stop')

        return message_count / true_duration

    async def receive(self, topic: Topic) -> None:
        message_count = 0
        last_count = None
        avg_latency = 0.
        stop_signal = asyncio.locks.Event()

        async def handle_warmup(topic_):
            pass

        async def handle_message(topic_, data_):
            nonlocal message_count, avg_latency

            now = time.time()
            send_time = data_[0]
            dt = now - send_time

            message_count += 1
            avg_latency += (dt - avg_latency) / message_count

        async def handle_stop(topic_):
            print(f'[{self.node}] Received stop signal')
            stop_signal.set()

        await self.node.listen('warmup', handle_warmup)
        await self.node.listen(topic, handle_message)
        await self.node.listen('stop', handle_stop)

        sleep_time = 1.
        while not stop_signal.is_set():
            await asyncio.sleep(sleep_time)

            if message_count != last_count:
                mps = (message_count - (last_count or 0)) / sleep_time
                print(f'[{self.node}] Received message count: {message_count}; mps={round(mps)}')
                last_count = message_count

                print(f'[{self.node}] Avg latency: {avg_latency}s')


def add_speedtest_command(subparsers) -> None:
    parser: ArgumentParser = subparsers.add_parser(
        'speedtest',
        description='Run a speed test between two nodes.',
        help='Run a speed test',
    )

    parser.add_argument(
        'role', choices=('send', 'recv'),
        help='Role of the node: sender or receiver.',
    )
    add_coordinator_arg(parser)
    add_authkey_arg(parser)
    parser.add_argument(
        '--seconds', type=float, default=10.,
        help='How long to run the speed test in seconds. Default: %(default)s',
    )
    parser.add_argument(
        '--topic', default='t',
        help='Topic to send/receive on. Default: %(default)s',
    )
    parser.add_argument(
        '--message-size', type=int, default=0,
        help='Size of the message to send in bytes. Default: 0 (no data).',
    )
    parser.add_argument(
        '--codec',
        choices=('pickle', 'json', 'msgpack'),
        default='pickle',
        help='Codec to use for encoding/decoding messages. Default: %(default)s.',
    )
    parser.add_argument(
        '--enable-load-balancer', action='store_true',
        help='Enable the default load balancer. It is disabled for speed testing by default.',
    )
    parser.add_argument(
        '--disable-unix', action='store_true',
        help='Disable Unix domain sockets for inter-node connections.',
    )
    parser.add_argument(
        '--disable-tcp', action='store_true',
        help='Disable TCP sockets for inter-node connections.',
    )



import asyncio
import logging

from rosy import build_node
from rosy.types import Topic


async def main():
    logging.basicConfig(level='WARNING')

    node = await build_node('topic_listener')
    await node.listen('some-topic', callback)
    await node.forever()


async def callback(topic: Topic, message: str, name: str = None):
    print(f'Received "{message} {name}" on topic={topic}')


if __name__ == '__main__':
    asyncio.run(main())



import argparse
import asyncio
import logging
from argparse import ArgumentParser, Namespace
from datetime import datetime

from rosy import build_node_from_args
from rosy.argparse import add_node_args
from rosy.cli.topic.send import parse_args_and_kwargs
from rosy.cli.topic.utils import print_args_and_kwargs
from rosy.cli.utils import add_log_arg


async def call_main(args: Namespace):
    logging.basicConfig(level=args.log)

    node = await build_node_from_args(args=args)

    service = node.get_service(args.service)

    async def call_once():
        if not args.no_wait and not await service.has_providers():
            print(f'Waiting for providers...')
            await service.wait_for_provider()

        service_args, service_kwargs = parse_args_and_kwargs(args.args)

        now = datetime.now()
        print(f'[{now}]')
        print(f'Calling service={args.service!r}')
        print_args_and_kwargs(service_args, service_kwargs)

        response = await service.call(*service_args, **service_kwargs)
        print(f'Response: {response!r}')
        print()

    if args.interval < 0:
        await call_once()
        return

    while True:
        await call_once()
        await asyncio.sleep(args.interval)


def add_call_command(subparsers) -> None:
    parser: ArgumentParser = subparsers.add_parser(
        'call',
        description='''
Start a node that calls a service with the given arguments.

---

Simple example:
    rosy service call my-service "'hello world'" 42 pi=3.14

This will call service `my-service` with args `'hello world'` and 42, and keyword arg `pi=3.14`.

---

Intermediate example:
    rosy service call my-service "{'key': 'value', 'data': [1, 2, 3]}"

This will call the service with a dictionary as an argument.

---

Advanced example:
    rosy service call my-service "call:myproj.MyData('data', pi=3.14)" "send_time=call:time.time()"

This will call the specified class/function to populate the arguments.

---
'''.strip(),
        help='call a service',
        formatter_class=argparse.RawTextHelpFormatter,
    )

    parser.add_argument(
        'service',
        help='The service to call.',
    )

    parser.add_argument(
        'args',
        nargs='*',
        metavar='arg/kwarg',
        help='''
Arg(s) and/or kwarg(s) to call the service with.
Each arg must be a valid Python expression,
or follow the format:
`call:module.callable(*args, **kwargs)`
'''.strip(),
    )

    parser.add_argument(
        '--interval', '-i',
        default=-1,
        type=float,
        help='The interval in seconds to make calls. A value < 0 will '
             'cause the call to be made only once. Default: %(default)s',
    )

    parser.add_argument(
        '--no-wait',
        action='store_true',
        help='Send the message without waiting for any listeners',
    )

    add_log_arg(parser)

    add_node_args(
        parser,
        default_node_name='rosy service call',
    )



import asyncio
from abc import ABC, abstractmethod
from asyncio import Event
from collections.abc import Awaitable, Callable
from dataclasses import dataclass, field
from typing import Any

from rosy.objectio import ObjectIO

RequestId = int
Data = Any
RequestHandler = Callable[[Data], Awaitable[Data]]
MessageHandler = Callable[[Data], Awaitable[None]]


@dataclass
class Request:
    id: RequestId
    data: Data


@dataclass
class Response:
    id: RequestId
    data: Data


@dataclass
class Message:
    data: Data


@dataclass
class PendingResponse:
    data: Data = None
    received: Event = field(default_factory=Event)


class RPC(ABC):
    request_handler: RequestHandler = None
    message_handler: MessageHandler = None

    async def start(self) -> None:
        asyncio.create_task(self.run_forever())

    async def run_forever(self) -> None:
        pass

    @abstractmethod
    async def send_request(self, data: Data) -> Data:
        ...

    @abstractmethod
    async def send_message(self, data: Data) -> None:
        ...

    async def handle_request(self, data: Data) -> Data:
        if self.request_handler is not None:
            return await self.request_handler(data)
        else:
            raise NotImplementedError()

    async def handle_message(self, data: Data) -> None:
        if self.message_handler is not None:
            await self.message_handler(data)
        else:
            raise NotImplementedError()


class ObjectIORPC(RPC):
    def __init__(self, obj_io: ObjectIO[Any]):
        self.obj_io = obj_io

        self._request_id_counter: RequestId = -1
        self._pending_responses: dict[RequestId, PendingResponse] = {}

    async def run_forever(self) -> None:
        async for obj in self.obj_io.reader:
            if isinstance(obj, Request):
                await self._handle_received_request(obj)
            elif isinstance(obj, Response):
                await self._handle_received_response(obj)
            elif isinstance(obj, Message):
                await self.handle_message(obj.data)
            else:
                raise Exception(f'Received invalid object of type={type(obj)}')

    async def _handle_received_request(self, request: Request) -> None:
        response_data = await self.handle_request(request.data)
        response = Response(id=request.id, data=response_data)
        await self.obj_io.writer.write(response)

    async def _handle_received_response(self, response: Response) -> None:
        pending_response = self._pending_responses.pop(response.id)
        pending_response.data = response.data
        pending_response.received.set()

    async def send_request(self, data: Data) -> Data:
        request = Request(id=self._get_request_id(), data=data)

        pending_response = PendingResponse()
        self._pending_responses[request.id] = pending_response

        await self.obj_io.writer.write(request)
        await pending_response.received.wait()

        return pending_response.data

    def _get_request_id(self) -> int:
        self._request_id_counter += 1
        return self._request_id_counter

    async def send_message(self, data: Data) -> None:
        await self.obj_io.writer.write(Message(data))



import asyncio
import logging

from rosy import build_node


async def main():
    logging.basicConfig(level='WARNING')

    node = await build_node(name='service_caller')
    print('Calculating 2 * 2...')
    result = await node.call('multiply', 2, 2)
    print(f'Result: {result}')


if __name__ == '__main__':
    asyncio.run(main())



import asyncio
import pickle
from argparse import Namespace
from datetime import datetime
from pathlib import Path

from rosy import Node


async def record(node: Node, args: Namespace) -> None:
    bag_file_path = args.output or get_bag_file_path()

    with open(bag_file_path, 'wb') as bag_file:
        message_counter = 0

        async def callback(topic, *args_, **kwargs_) -> None:
            nonlocal message_counter

            now = datetime.now()
            pickle.dump((now, topic, args_, kwargs_), bag_file)

            message_counter += 1
            if not args.no_dots:
                print('.', end='', flush=True)

        for topic in args.topics:
            await node.listen(topic, callback)

        print(f'Recording topics to "{bag_file_path}":')
        for topic in args.topics:
            print(f'- {topic}')

        try:
            await node.forever()
        except asyncio.CancelledError:
            print('\nRecording stopped.')

    if message_counter == 0:
        print('No messages recorded.')
        bag_file_path.unlink()
    else:
        print(f'Recorded {message_counter} messages to "{bag_file_path}".')


def add_record_args(subparsers) -> None:
    parser = subparsers.add_parser('record', help='Record messages to file')

    parser.add_argument(
        '--output', '-o',
        type=Path,
        help='Output file path. Default: record_<YYYY-MM-DD>-<HH-MM-SS>.bag',
    )

    parser.add_argument(
        '--no-dots',
        action='store_true',
        help='Disable logging of "." when a new message is received.'
    )

    parser.add_argument(
        'topics',
        nargs='+',
        help='Topics to record.',
    )


def get_bag_file_path() -> Path:
    now = datetime.now()
    now = now.strftime("%Y-%m-%d-%H-%M-%S")
    return Path(f'record_{now}.bag')



import logging

from rosy.asyncio import LockableWriter
from rosy.node.codec import NodeMessageCodec
from rosy.node.service.handlermanager import ServiceHandlerManager
from rosy.node.service.types import ServiceRequest, ServiceResponse

logger = logging.getLogger(__name__)


class ServiceRequestHandler:
    def __init__(
            self,
            service_handler_manager: ServiceHandlerManager,
            node_message_codec: NodeMessageCodec,
    ):
        self.service_handler_manager = service_handler_manager
        self.node_message_codec = node_message_codec

    async def handle_request(
            self,
            request: ServiceRequest,
            writer: LockableWriter,
    ) -> None:
        handler = self.service_handler_manager.get_callback(request.service)

        result, error = None, None

        if handler is None:
            logger.warning(
                f'Received service request for service={request.service!r} '
                f'but no handler is registered for it.'
            )

            error = f'service={request.service!r} is not provided by this node'
        else:
            try:
                result = await handler(request.service, *request.args, **request.kwargs)
            except Exception as e:
                logger.exception(
                    f'Error handling service request={request}',
                    exc_info=e,
                )
                error = repr(e)

        response = ServiceResponse(request.id, result, error)

        async with writer:
            await self.node_message_codec.encode_service_response(
                writer,
                response,
            )



from collections.abc import Sequence
from dataclasses import dataclass
from typing import Any, Awaitable, Callable, Protocol

Data = Any

Topic = str
TopicCallback = Callable[[Topic], Awaitable[None]] | Callable[[Topic, ...], Awaitable[None]]

Service = str
ServiceCallback = Callable[[Service], Awaitable[Data]] | Callable[[Service, ...], Awaitable[Data]]

Host = str
ServerHost = Host | Sequence[Host] | None
Port = int


@dataclass
class Endpoint:
    host: Host
    port: Port

    def __str__(self):
        return f'{self.host}:{self.port}'


class Buffer(Protocol):
    """Not available in std lib until Python 3.12."""

    def __buffer__(self, *args, **kwargs):
        ...  # pragma: no cover



import logging
from asyncio import Lock, open_connection, open_unix_connection
from collections.abc import Iterable
from typing import NamedTuple

from rosy.asyncio import LockableWriter, Reader, Writer
from rosy.authentication import Authenticator
from rosy.network import get_hostname
from rosy.node.loadbalancing import ServiceLoadBalancer, TopicLoadBalancer
from rosy.node.topology import MeshTopologyManager
from rosy.specs import ConnectionSpec, IpConnectionSpec, MeshNodeSpec, NodeId, UnixConnectionSpec
from rosy.types import Host, Service, Topic

logger = logging.getLogger(__name__)


class PeerConnection(NamedTuple):
    reader: Reader
    writer: LockableWriter

    async def close(self) -> None:
        self.writer.close()
        await self.writer.wait_closed()

    def is_closing(self) -> bool:
        return self.writer.is_closing()


class PeerConnectionBuilder:
    def __init__(self, authenticator: Authenticator, host: Host = None):
        self.authenticator = authenticator
        self.host = host or get_hostname()

    async def build(self, conn_specs: Iterable[ConnectionSpec]) -> tuple[Reader, Writer]:
        reader_writer = None
        for conn_spec in conn_specs:
            try:
                reader_writer = await self._get_connection(conn_spec)
            except (ConnectionError, IOError) as e:
                logger.error(f'Error connecting to {conn_spec}: {e!r}')
                continue

            if reader_writer is not None:
                break

        if reader_writer is None:
            raise ConnectionError('Could not connect to any connection spec')

        reader, writer = reader_writer
        await self.authenticator.authenticate(reader, writer)

        return reader, writer

    async def _get_connection(self, conn_spec: ConnectionSpec) -> tuple[Reader, Writer] | None:
        if isinstance(conn_spec, IpConnectionSpec):
            return await open_connection(
                host=conn_spec.host,
                port=conn_spec.port,
                family=conn_spec.family,
            )
        elif isinstance(conn_spec, UnixConnectionSpec):
            if conn_spec.host != self.host:
                return None

            return await open_unix_connection(path=conn_spec.path)
        else:
            raise ValueError(f'Unrecognized connection spec: {conn_spec}')


class PeerConnectionManager:
    def __init__(self, conn_builder: PeerConnectionBuilder):
        self.conn_builder = conn_builder

        self._connections: dict[NodeId, PeerConnection] = {}
        self._connections_lock = Lock()

    async def get_connection(self, node: MeshNodeSpec) -> PeerConnection:
        async with self._connections_lock:
            connection = await self._get_cached_connection(node)
            if connection:
                return connection

            logger.debug(f'Connecting to node: {node.id}')
            reader, writer = await self.conn_builder.build(node.connection_specs)

            writer = LockableWriter(writer)

            connection = PeerConnection(reader, writer)
            self._connections[node.id] = connection
            return connection

    async def _get_cached_connection(self, node: MeshNodeSpec) -> PeerConnection | None:
        connection = self._connections.get(node.id, None)
        if not connection:
            return None

        if connection.is_closing():
            await self.close_connection(node)
            return None

        return connection

    async def close_connection(self, node: MeshNodeSpec) -> None:
        connection = self._connections.pop(node.id, None)
        if connection:
            await connection.close()


class PeerSelector:
    def __init__(
            self,
            topology_manager: MeshTopologyManager,
            topic_load_balancer: TopicLoadBalancer,
            service_load_balancer: ServiceLoadBalancer,
    ):
        self.topology_manager = topology_manager
        self.topic_load_balancer = topic_load_balancer
        self.service_load_balancer = service_load_balancer

    def get_nodes_for_topic(self, topic: Topic) -> list[MeshNodeSpec]:
        peers = self.topology_manager.get_nodes_listening_to_topic(topic)
        return self.topic_load_balancer.choose_nodes(peers, topic)

    def get_node_for_service(self, service: Service) -> MeshNodeSpec | None:
        peers = self.topology_manager.get_nodes_providing_service(service)
        return self.service_load_balancer.choose_node(peers, service)



import logging

from rosy.asyncio import log_error
from rosy.node.topic.listenermanager import TopicListenerManager
from rosy.node.topic.types import TopicMessage

logger = logging.getLogger(__name__)


class TopicMessageHandler:
    def __init__(
            self,
            listener_manager: TopicListenerManager,
    ):
        self.listener_manager = listener_manager

    async def handle_message(self, message: TopicMessage) -> None:
        callback = self.listener_manager.get_callback(message.topic)

        if callback:
            await log_error(callback(message.topic, *message.args, **message.kwargs))
        else:
            logger.warning(
                f'Received message for topic={message.topic!r} '
                f'but no listener is registered.'
            )



import asyncio
import logging

from rosy import build_node


async def main():
    logging.basicConfig(level='WARNING')

    node = await build_node(name='topic_sender')
    await node.send('some-topic', 'hello', name='world')


if __name__ == '__main__':
    asyncio.run(main())



import pickle
from abc import ABC, abstractmethod
from collections.abc import Sequence
from typing import Any, Generic, Literal, TypeVar

import msgpack
import orjson

from rosy.asyncio import BufferWriter, Reader, Writer
from rosy.utils import require

T = TypeVar('T')
K = TypeVar('K')
V = TypeVar('V')

ByteOrder = Literal['big', 'little']
DEFAULT_BYTE_ORDER: ByteOrder = 'little'

DEFAULT_MAX_BYTE_LENGTH: int = 8

DEFAULT_MAX_TOPIC_LENGTH: int = 256
DEFAULT_TOPIC_ENCODING: str = 'utf-8'


class Codec(Generic[T], ABC):
    @abstractmethod
    async def encode(self, writer: Writer, obj: T) -> None:
        ...  # pragma: no cover

    @abstractmethod
    async def decode(self, reader: Reader) -> T:
        ...  # pragma: no cover


class FixedLengthIntCodec(Codec[int]):
    def __init__(
            self,
            length: int,
            byte_order: ByteOrder = DEFAULT_BYTE_ORDER,
            signed: bool = False,
    ):
        self.length = length
        self.byte_order = byte_order
        self.signed = signed

    async def encode(self, writer: Writer, value: int) -> None:
        data = value.to_bytes(self.length, byteorder=self.byte_order, signed=self.signed)
        writer.write(data)

    async def decode(self, reader: Reader) -> int:
        data = await reader.readexactly(self.length)
        return int.from_bytes(data, byteorder=self.byte_order, signed=self.signed)


class VariableLengthIntCodec(Codec[int]):
    def __init__(
            self,
            max_byte_length: int = DEFAULT_MAX_BYTE_LENGTH,
            byte_order: ByteOrder = DEFAULT_BYTE_ORDER,
            signed: bool = False,
    ):
        require(
            0 < max_byte_length <= 255,
            f'max_byte_length must be in range (0, 255]; got {max_byte_length}.',
        )

        self.max_byte_length = max_byte_length
        self.byte_order = byte_order
        self.signed = signed

    async def encode(self, writer: Writer, value: int) -> None:
        int_byte_length = byte_length(value)

        if int_byte_length > self.max_byte_length:
            raise OverflowError(
                f'Computed byte_length={int_byte_length} is greater than '
                f'max_byte_length={self.max_byte_length}'
            )

        header = bytes([int_byte_length])

        if int_byte_length > 0:
            data = value.to_bytes(int_byte_length, byteorder=self.byte_order, signed=self.signed)
        else:
            data = None

        writer.write(header)
        if data:
            writer.write(data)

    async def decode(self, reader: Reader) -> int:
        header = await reader.readexactly(1)

        int_byte_length = header[0]
        if int_byte_length == 0:
            return 0

        require(
            int_byte_length <= self.max_byte_length,
            f'Received byte_length={int_byte_length} > max_byte_length={self.max_byte_length}',
        )

        data = await reader.readexactly(int_byte_length)
        return int.from_bytes(data, byteorder=self.byte_order, signed=self.signed)


class LengthPrefixedStringCodec(Codec[str]):
    def __init__(
            self,
            len_prefix_codec: Codec[int],
            encoding: str = 'utf-8',
    ):
        self.len_prefix_codec = len_prefix_codec
        self.encoding = encoding

    async def encode(self, writer: Writer, data: str) -> None:
        data = data.encode(encoding=self.encoding)
        await self.len_prefix_codec.encode(writer, len(data))
        if data:
            writer.write(data)

    async def decode(self, reader: Reader) -> str:
        length = await self.len_prefix_codec.decode(reader)
        if length == 0:
            return ''

        data = await reader.readexactly(length)
        return data.decode(encoding=self.encoding)


def byte_length(value: int) -> int:
    """Returns the number of bytes required to represent an integer."""
    return (value.bit_length() + 7) // 8


class SequenceCodec(Generic[T], Codec[Sequence[T]]):
    def __init__(
            self,
            len_header_codec: Codec[int],
            item_codec: Codec[T],
    ):
        self.len_header_codec = len_header_codec
        self.item_codec = item_codec

    async def encode(self, writer: Writer, sequence: Sequence[T]) -> None:
        await self.len_header_codec.encode(writer, len(sequence))

        for item in sequence:
            await self.item_codec.encode(writer, item)

    async def decode(self, reader: Reader) -> Sequence[T]:
        length = await self.len_header_codec.decode(reader)

        return [
            await self.item_codec.decode(reader)
            for _ in range(length)
        ]


class DictCodec(Generic[K, V], Codec[dict[K, V]]):
    def __init__(
            self,
            len_header_codec: Codec[int],
            key_codec: Codec[K],
            value_codec: Codec[V],
    ):
        self.len_header_codec = len_header_codec
        self.key_codec = key_codec
        self.value_codec = value_codec

    async def encode(self, writer: Writer, dict_: dict[K, V]) -> None:
        await self.len_header_codec.encode(writer, len(dict_))

        for key, value in dict_.items():
            await self.key_codec.encode(writer, key)
            await self.value_codec.encode(writer, value)

    async def decode(self, reader: Reader) -> dict[K, V]:
        length = await self.len_header_codec.decode(reader)

        result = {}
        for _ in range(length):
            key = await self.key_codec.decode(reader)
            value = await self.value_codec.decode(reader)
            result[key] = value

        return result


class PickleCodec(Codec[Any]):
    def __init__(
            self,
            protocol: int = pickle.HIGHEST_PROTOCOL,
            dump_kwargs: dict[str, Any] = None,
            load_kwargs: dict[str, Any] = None,
            len_header_bytes: int = 4,
            len_header_codec: Codec[int] = None,
    ):
        self.protocol = protocol
        self.dump_kwargs = dump_kwargs or {}
        self.load_kwargs = load_kwargs or {}
        self.len_header_codec = len_header_codec or FixedLengthIntCodec(len_header_bytes)

    async def encode(self, writer: Writer, obj: Any) -> None:
        buffer = BufferWriter()
        pickle.dump(obj, buffer, protocol=self.protocol, **self.dump_kwargs)

        await self.len_header_codec.encode(writer, len(buffer))
        writer.write(buffer)

    async def decode(self, reader: Reader) -> Any:
        data_len = await self.len_header_codec.decode(reader)
        data = await reader.readexactly(data_len)
        return pickle.loads(data, **self.load_kwargs)


pickle_codec = PickleCodec()
"""Pickle codec with default settings. Encoded data can be up to 4 GiB in size."""


class JsonCodec(Codec[Any]):
    def __init__(
            self,
            dumps_kwargs: dict[str, Any] = None,
            len_header_bytes: int = 4,
            len_header_codec: Codec[int] = None,
    ):
        self.dumps_kwargs = dumps_kwargs or {}
        self.len_header_codec = len_header_codec or FixedLengthIntCodec(len_header_bytes)

    async def encode(self, writer: Writer, obj: T) -> None:
        data = orjson.dumps(obj, **self.dumps_kwargs)
        await self.len_header_codec.encode(writer, len(data))
        writer.write(data)

    async def decode(self, reader: Reader) -> T:
        data_len = await self.len_header_codec.decode(reader)
        data = await reader.readexactly(data_len)
        return orjson.loads(data)


json_codec = JsonCodec()
"""JSON codec with default settings. Encoded data can be up to 4 GiB in size."""

MsgpackTypes = None | bool | int | float | str | bytes | bytearray | list | tuple | dict


class MsgpackCodec(Codec[MsgpackTypes]):
    def __init__(
            self,
            pack_kwargs: dict[str, Any] = None,
            unpack_kwargs: dict[str, Any] = None,
            len_header_bytes: int = 4,
            len_header_codec: Codec[int] = None,
    ):
        self.pack_kwargs = pack_kwargs or {}
        self.unpack_kwargs = unpack_kwargs or {}
        self.len_header_codec = len_header_codec or FixedLengthIntCodec(len_header_bytes)

    async def encode(self, writer: Writer, obj: MsgpackTypes) -> None:
        data = msgpack.packb(obj, **self.pack_kwargs)
        await self.len_header_codec.encode(writer, len(data))
        writer.write(data)

    async def decode(self, reader: Reader) -> MsgpackTypes:
        data_len = await self.len_header_codec.decode(reader)
        data = await reader.readexactly(data_len)
        return msgpack.unpackb(data, **self.unpack_kwargs)


msgpack_codec = MsgpackCodec()



import asyncio
from argparse import Namespace
from datetime import datetime
from pathlib import Path

from rosy.cli.bag.file import get_bag_file_messages, get_most_recent_bag_file_path
from rosy import Node
from rosy.utils import require


async def play(node: Node, args: Namespace) -> None:
    bag_file_path = args.input or get_most_recent_bag_file_path()

    print(f'Playing back messages from "{bag_file_path}"...')

    first_instant = None
    first_sent_instant = datetime.now()

    for instant, topic, args_, kwargs_ in get_bag_file_messages(bag_file_path):
        if not first_instant:
            first_instant = instant
        elif not args.immediate:
            await _wait_for_next_send(
                first_instant,
                first_sent_instant,
                instant,
                args.rate,
            )

        await node.send(topic, *args_, **kwargs_)

        if not args.no_log:
            if args.no_log_data:
                print(f'[{instant}] {topic}')
            else:
                print(f'[{instant}] {topic}: args={args_}; kwargs={kwargs_}')


def add_play_args(subparsers) -> None:
    parser = subparsers.add_parser('play', help='Playback recorded messages from file')

    parser.add_argument(
        '--input', '-i',
        type=Path,
        help='Input file path. Default: The most recent '
             'record_*.bag file in the current directory.',
    )

    parser.add_argument(
        '--rate', '-r',
        type=float,
        default=1.0,
        help='Playback rate. Default: %(default)s',
    )

    parser.add_argument(
        '--immediate',
        action='store_true',
        help='Send messages immediately without waiting for the next send time.',
    )

    parser.add_argument(
        '--no-log',
        action='store_true',
        help='Do not log sent messages to the console.',
    )

    parser.add_argument(
        '--no-log-data',
        action='store_true',
        help='Do not log data from sent messages to the console.',
    )


async def _wait_for_next_send(
        first_instant: datetime,
        first_sent_instant: datetime,
        instant: datetime,
        rate: float,
) -> None:
    require(rate > 0, f'Rate must be greater than 0; got {rate}')

    dt = (instant - first_instant) / rate
    send_instant = first_sent_instant + dt
    now = datetime.now()
    wait_time = (send_instant - now).total_seconds()

    await asyncio.sleep(wait_time)



from argparse import ArgumentParser, Namespace

from rosy import build_node_from_args
from rosy.argparse import add_node_args
from rosy.cli.bag.info import add_info_args, display_info
from rosy.cli.bag.play import add_play_args, play
from rosy.cli.bag.record import add_record_args, record


async def bag_main(args: Namespace):
    print(args)
    if args.bag_command == 'record':
        node = await build_node_from_args(args=args)
        await record(node, args)
    elif args.bag_command == 'play':
        node = await build_node_from_args(args=args)
        await play(node, args)
    elif args.bag_command == 'info':
        await display_info(args)
    else:
        raise ValueError(f'Unknown command: {args.bag_command}')


def add_bag_command(subparsers) -> None:
    parser: ArgumentParser = subparsers.add_parser(
        'bag',
        description='Tool for recording and playing back topic messages. '
                    'Based on the `ros2 bag` ROS command line tool.',
        help='Record and play back topic messages',
    )

    add_node_args(
        parser,
        default_node_name='rosy bag',
    )

    subparsers = parser.add_subparsers(
        title='commands',
        dest='bag_command',
        required=True,
    )

    add_record_args(subparsers)
    add_play_args(subparsers)
    add_info_args(subparsers)



import asyncio
import logging
import tempfile
from abc import ABC, abstractmethod
from asyncio import Server, StreamReader, StreamWriter
from collections.abc import Awaitable, Callable, Iterable

from rosy.asyncio import Reader, Writer
from rosy.specs import ConnectionSpec, IpConnectionSpec, UnixConnectionSpec
from rosy.types import Host, Port, ServerHost

ClientConnectedCallback = Callable[[Reader, Writer], Awaitable[None]]

logger = logging.getLogger(__name__)


class ServerProvider(ABC):
    @abstractmethod
    async def start_server(
            self,
            client_connected_cb,
    ) -> tuple[Server, list[ConnectionSpec]]:
        """Raises ``UnsupportedProviderError`` if not supported on the system."""
        ...  # pragma: no cover


class TcpServerProvider(ServerProvider):
    """Starts a TCP server on the first available port."""

    def __init__(
            self,
            server_host: ServerHost,
            client_host: Host,
            port: Port = 0,
            **kwargs,
    ):
        """
        Args:
            server_host:
                The interface(s) that the server will listen on.
            client_host:
                The host that clients will use to connect to the server.
            port:
                The port to start the server on. If set to 0, the server will
                choose an available port automatically.
            kwargs:
                Additional keyword arguments will be passed to the
                ``asyncio.start_server`` call.
        """

        self.server_host = server_host
        self.client_host = client_host
        self.port = port
        self.kwargs = kwargs

    async def start_server(
            self,
            client_connected_cb,
    ) -> tuple[Server, list[ConnectionSpec]]:
        server = await asyncio.start_server(
            client_connected_cb,
            host=self.server_host,
            port=self.port,
            **self.kwargs,
        )

        conn_specs = [
            IpConnectionSpec(
                self.client_host,
                port=socket.getsockname()[1],
                family=socket.family,
            )
            for socket in server.sockets
        ]

        # Sort IPv4 before IPv6
        conn_specs.sort(key=lambda spec: spec.family)

        return server, conn_specs


class TmpUnixServerProvider(ServerProvider):
    """Starts a Unix server on a tmp file."""

    def __init__(
            self,
            prefix: str | None = 'rosy-node-server.',
            suffix: str | None = '.sock',
            dir=None,
            **kwargs,
    ):
        """
        Args:
            prefix:
                The prefix for the temporary Unix socket file.
            suffix:
                The suffix for the temporary Unix socket file.
            dir:
                The directory to create the temporary Unix socket file in.
                If not set, the system's default temporary directory will be used.
            kwargs:
                Additional keyword arguments will be passed to the
                ``asyncio.start_unix_server`` call.
        """

        self.prefix = prefix
        self.suffix = suffix
        self.dir = dir
        self.kwargs = kwargs

    async def start_server(
            self,
            client_connected_cb,
    ) -> tuple[Server, list[ConnectionSpec]]:
        with tempfile.NamedTemporaryFile(
                prefix=self.prefix,
                suffix=self.suffix,
                dir=self.dir,
        ) as file:
            path = file.name

        try:
            server = await asyncio.start_unix_server(client_connected_cb, path=path, **self.kwargs)
        except NotImplementedError as e:
            raise UnsupportedProviderError(self, repr(e))

        conn_spec = UnixConnectionSpec(path=path)

        return server, [conn_spec]


class ServersManager:
    def __init__(
            self,
            server_providers: Iterable[ServerProvider],
            client_connected_cb: ClientConnectedCallback,
    ):
        self.server_providers = server_providers
        self.client_connected_cb = client_connected_cb

        self._connection_specs: list[ConnectionSpec] = []

    @property
    def connection_specs(self) -> list[ConnectionSpec]:
        return list(self._connection_specs)

    async def start_servers(self) -> None:
        if self._connection_specs:
            raise RuntimeError('Servers have already been started.')

        client_connected_cb = _close_on_return(self.client_connected_cb)

        for provider in self.server_providers:
            try:
                server, connection_specs = await provider.start_server(client_connected_cb)
            except UnsupportedProviderError as e:
                logger.exception(f'Failed to start server using provider={provider}', exc_info=e)
            else:
                logger.debug(f'Started node server with connection_specs={connection_specs}')
                self._connection_specs.extend(connection_specs)

        if not self._connection_specs:
            raise RuntimeError('Unable to start any server with the given server providers.')


class UnsupportedProviderError(Exception):
    """Raised when trying to start a server from an unsupported server provider."""

    def __init__(self, provider: ServerProvider, message: str):
        super().__init__(
            f'Unsupported server provider {provider.__class__}: {message}'
        )


def _close_on_return(
        callback: ClientConnectedCallback,
) -> Callable[[StreamReader, StreamWriter], Awaitable[None]]:
    async def wrapped_callback(reader: StreamReader, writer: StreamWriter) -> None:
        try:
            await callback(reader, writer)
        except Exception as e:
            logger.exception('Error in client connected callback', exc_info=e)
        finally:
            writer.close()
            await writer.wait_closed()

    return wrapped_callback



from rosy.asyncio import log_error, many
from rosy.node.codec import NodeMessageCodec
from rosy.node.peer import PeerConnectionManager, PeerSelector
from rosy.node.topic.types import TopicMessage
from rosy.node.types import Args, KWArgs
from rosy.specs import MeshNodeSpec
from rosy.types import Buffer, Topic


class TopicSender:
    def __init__(
            self,
            peer_selector: PeerSelector,
            connection_manager: PeerConnectionManager,
            node_message_codec: NodeMessageCodec,
    ):
        self.peer_selector = peer_selector
        self.connection_manager = connection_manager
        self.node_message_codec = node_message_codec

    async def send(self, topic: Topic, args: Args, kwargs: KWArgs) -> None:
        # TODO handle case of self-sending more efficiently

        nodes = self.peer_selector.get_nodes_for_topic(topic)
        if not nodes:
            return

        message = TopicMessage(topic, args, kwargs)
        data = await self.node_message_codec.encode_topic_message(message)

        await many([
            log_error(self._send_to_one(n, data))
            for n in nodes
        ])

    async def _send_to_one(self, node: MeshNodeSpec, data: Buffer) -> None:
        connection = await self.connection_manager.get_connection(node)

        async with connection.writer as writer:
            writer.write(data)
            await writer.drain()



import argparse
import asyncio
import sys

from rosy.cli.bag.main import add_bag_command, bag_main
from rosy.cli.coordinator import add_coordinator_command, coordinator_main
from rosy.cli.launch.main import add_launch_command, launch_main
from rosy.cli.node.main import add_node_command, node_main
from rosy.cli.service.main import add_service_command, service_main
from rosy.cli.speedtest import add_speedtest_command, speedtest_main
from rosy.cli.topic.main import add_topic_command, topic_main

_command_to_main = {
    'coordinator': coordinator_main,
    'bag': bag_main,
    'launch': launch_main,
    'node': node_main,
    'topic': topic_main,
    'service': service_main,
    'speedtest': speedtest_main,
}

_add_command_functions = [
    add_coordinator_command,
    add_bag_command,
    add_launch_command,
    add_node_command,
    add_topic_command,
    add_service_command,
    add_speedtest_command,
]


def main() -> None:
    try:
        asyncio.run(_main())
    except KeyboardInterrupt:
        pass


async def _main():
    parser = get_arg_parser()

    args = sys.argv[1:] or ['coordinator']
    args = parser.parse_args(args)

    command_main = _command_to_main.get(args.command)
    await command_main(args)


def get_arg_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description='rosy CLI',
    )

    subparsers = parser.add_subparsers(
        title='commands',
        dest='command',
        required=True,
    )

    for add_command in _add_command_functions:
        add_command(subparsers)

    return parser



import asyncio
import logging
from asyncio import Future
from collections.abc import Iterator
from contextlib import contextmanager
from weakref import WeakKeyDictionary

from rosy.asyncio import Reader
from rosy.node.codec import NodeMessageCodec
from rosy.node.peer import PeerConnectionManager, PeerSelector
from rosy.node.service.types import RequestId, ServiceRequest, ServiceResponse
from rosy.node.types import Args, KWArgs
from rosy.types import Data

logger = logging.getLogger(__name__)


class ServiceCaller:
    def __init__(
            self,
            peer_selector: PeerSelector,
            connection_manager: PeerConnectionManager,
            node_message_codec: NodeMessageCodec,
            max_request_ids: int,
    ):
        self.peer_selector = peer_selector
        self.connection_selector = connection_manager
        self.node_message_codec = node_message_codec
        self.max_request_ids = max_request_ids

        self._next_request_id: RequestId = 0
        self._response_futures: WeakKeyDictionary[
            Reader,
            dict[RequestId, Future]
        ] = WeakKeyDictionary()

    async def call(self, service: str, args: Args, kwargs: KWArgs) -> Data:
        node = self.peer_selector.get_node_for_service(service)
        if node is None:
            raise ValueError(f'No node hosting service={service!r}')

        connection = await self.connection_selector.get_connection(node)

        self._start_response_handler(connection.reader)

        with self._get_request_id_and_response_future(connection.reader) as (request_id, response_future):
            request = ServiceRequest(request_id, service, args, kwargs)
            request = await self.node_message_codec.encode_service_request(request)

            async with connection.writer as writer:
                writer.write(request)
                await writer.drain()

            response: ServiceResponse = await response_future

        if response.error:
            raise ServiceResponseError(response.error)

        return response.result

    @contextmanager
    def _get_request_id_and_response_future(
            self,
            reader: Reader,
    ) -> Iterator[tuple[RequestId, Future]]:
        request_id = self._get_new_request_id(reader)
        response_future = Future()
        self._response_futures[reader][request_id] = response_future

        yield request_id, response_future

        futures = self._response_futures.get(reader)
        if futures:
            futures.pop(request_id, None)

    def _get_new_request_id(self, reader: Reader) -> RequestId:
        request_id = self._find_next_available_request_id(reader)
        self._inc_next_request_id()
        return request_id

    def _find_next_available_request_id(self, reader: Reader) -> RequestId:
        response_futures = self._response_futures[reader]
        for _ in range(self.max_request_ids):
            if self._next_request_id not in response_futures:
                return self._next_request_id

            self._inc_next_request_id()

        raise ServiceRequestError(
            f'All {self.max_request_ids} request IDs are in use for reader={reader!r}'
        )

    def _inc_next_request_id(self) -> None:
        self._next_request_id = (self._next_request_id + 1) % self.max_request_ids

    def _start_response_handler(self, reader: Reader) -> None:
        if reader not in self._response_futures:
            self._response_futures[reader] = {}
            asyncio.create_task(self._response_handler(reader), name='ServiceResponseHandler')

    async def _response_handler(self, reader: Reader) -> None:
        try:
            while True:
                await self._handle_one_response(reader)
        finally:
            self._fail_pending_response_futures_for(reader)
            self._response_futures.pop(reader)

    async def _handle_one_response(self, reader: Reader) -> None:
        response = await self.node_message_codec.decode_service_response(reader)

        response_future = self._response_futures[reader].get(response.id)
        if response_future is None:
            logger.warning(
                f'Received response for unknown request '
                f'id={response.id} on reader={reader!r}'
            )
            return

        if not response_future.done():
            response_future.set_result(response)

    def _fail_pending_response_futures_for(self, reader: Reader) -> None:
        error = ServiceResponseError(f'Reader {reader!r} was closed before response was received')
        response_futures = self._response_futures[reader].values()
        for response_future in response_futures:
            if not response_future.done():
                response_future.set_exception(error)


class ServiceRequestError(Exception):
    pass


class ServiceResponseError(Exception):
    pass



import asyncio
import logging

from rosy import build_node
from rosy.types import Service


async def main():
    logging.basicConfig(level='WARNING')

    node = await build_node('service_provider')
    await node.add_service('multiply', multiply)
    await node.forever()


async def multiply(service: Service, a: int, b: int) -> int:
    return a * b


if __name__ == '__main__':
    asyncio.run(main())



from argparse import ArgumentParser, Namespace

from rosy.cli.service.call import add_call_command, call_main
from rosy.cli.service.list import add_list_command, list_main


async def service_main(args: Namespace) -> None:
    if args.service_command == 'list':
        await list_main(args)
    elif args.service_command == 'call':
        await call_main(args)
    else:
        raise ValueError(f'Unknown service command: {args.service_command}')


def add_service_command(subparsers) -> None:
    parser: ArgumentParser = subparsers.add_parser(
        'service',
        description='Service commands.',
        help='Service commands like list and call',
    )

    subparsers = parser.add_subparsers(
        title='commands',
        dest='service_command',
        required=True,
    )

    add_list_command(subparsers)
    add_call_command(subparsers)



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Generic, Literal, TypeVar

from rosy.asyncio import Reader, Writer
from rosy.codec import Codec, pickle_codec

DEFAULT_TOPIC_ENCODING: str = 'utf-8'

ByteOrder = Literal['big', 'little']
DEFAULT_BYTE_ORDER: ByteOrder = 'little'

DEFAULT_MAX_HEADER_LEN: int = 8

T = TypeVar('T')


class ObjectReader(Generic[T], ABC):
    def __aiter__(self):
        return self

    async def __anext__(self) -> T:
        return await self.read()

    @abstractmethod
    async def read(self) -> T:
        ...


class ObjectWriter(Generic[T], ABC):
    @abstractmethod
    async def write(self, obj: T) -> None:
        ...


@dataclass
class ObjectIO(Generic[T]):
    reader: ObjectReader[T]
    writer: ObjectWriter[T]


class CodecObjectReader(ObjectReader[T]):
    def __init__(
            self,
            reader: Reader,
            codec: Codec[T] = pickle_codec,
    ):
        self.reader = reader
        self.codec = codec

    async def read(self) -> T:
        return await self.codec.decode(self.reader)


class CodecObjectWriter(ObjectWriter[T]):
    def __init__(
            self,
            writer: Writer,
            codec: Codec[T] = pickle_codec,
    ):
        self.writer = writer
        self.codec = codec

    async def write(self, obj: T) -> None:
        await self.codec.encode(self.writer, obj)
        await self.writer.drain()



import asyncio
import logging

from rosy.asyncio import LockableWriter, Reader, Writer
from rosy.authentication import Authenticator
from rosy.node.codec import NodeMessageCodec
from rosy.node.service.requesthandler import ServiceRequestHandler
from rosy.node.service.types import ServiceRequest
from rosy.node.topic.messagehandler import TopicMessageHandler
from rosy.node.topic.types import TopicMessage

logger = logging.getLogger(__name__)


class ClientHandler:
    def __init__(
            self,
            authenticator: Authenticator,
            node_message_codec: NodeMessageCodec,
            topic_message_handler: TopicMessageHandler,
            service_request_handler: ServiceRequestHandler,
    ):
        self.authenticator = authenticator
        self.node_message_codec = node_message_codec
        self.topic_message_handler = topic_message_handler
        self.service_request_handler = service_request_handler

    async def handle_client(self, reader: Reader, writer: Writer) -> None:
        peer_name = writer.get_extra_info('peername') or writer.get_extra_info('sockname')
        logger.debug(f'New connection from: {peer_name}')

        await self.authenticator.authenticate(reader, writer)
        writer = LockableWriter(writer)

        while True:
            try:
                obj = await self.node_message_codec.decode_topic_message_or_service_request(reader)
            except EOFError:
                logger.debug(f'Closed connection from: {peer_name}')
                return

            if isinstance(obj, TopicMessage):
                await self.topic_message_handler.handle_message(obj)
            elif isinstance(obj, ServiceRequest):
                asyncio.create_task(
                    self.service_request_handler.handle_request(obj, writer),
                    name=f'Handle service request {obj.id} from {peer_name}',
                )
            else:
                raise RuntimeError('Unreachable code')



from rosy.types import Port, ServerHost

DEFAULT_COORDINATOR_HOST: ServerHost = ''
"""Default coordinator host is empty string, which means all available interfaces."""

DEFAULT_COORDINATOR_PORT: Port = 7679
"""Default coordinator port is 7679, which spells "ROSY" on a phone keypad :)"""



from typing import Callable, Generic, TypeVar

K = TypeVar('K')
C = TypeVar('C', bound=Callable)


class CallbackManager(Generic[K, C]):
    def __init__(self):
        self._handlers: dict[K, C] = {}

    @property
    def keys(self) -> set[K]:
        return set(self._handlers.keys())

    def get_callback(self, key: K) -> Callable | None:
        return self._handlers.get(key)

    def set_callback(self, key: K, callback: Callable) -> None:
        self._handlers[key] = callback

    def remove_callback(self, key: K) -> Callable | None:
        return self._handlers.pop(key, None)



from rosy.asyncio import BufferWriter, Reader, Writer
from rosy.codec import Codec
from rosy.node.service.types import ServiceRequest, ServiceResponse
from rosy.node.topic.types import TopicMessage
from rosy.types import Buffer
from rosy.utils import require


class NodeMessageCodec:
    def __init__(
            self,
            topic_message_codec: Codec[TopicMessage],
            service_request_codec: Codec[ServiceRequest],
            service_response_codec: Codec[ServiceResponse],
            topic_message_prefix: bytes = b't',
            service_request_prefix: bytes = b's',
    ):
        require(len(topic_message_prefix) == 1, 'Topic message prefix must be a single byte')
        require(len(service_request_prefix) == 1, 'Service request prefix must be a single byte')

        self.topic_message_codec = topic_message_codec
        self.service_request_codec = service_request_codec
        self.service_response_codec = service_response_codec
        self.topic_message_prefix = topic_message_prefix
        self.service_request_prefix = service_request_prefix

    async def encode_topic_message(self, message: TopicMessage) -> Buffer:
        buffer = BufferWriter()
        buffer.write(self.topic_message_prefix)
        await self.topic_message_codec.encode(buffer, message)
        return buffer

    async def encode_service_request(self, request: ServiceRequest) -> Buffer:
        buffer = BufferWriter()
        buffer.write(self.service_request_prefix)
        await self.service_request_codec.encode(buffer, request)
        return buffer

    async def encode_service_response(
            self,
            writer: Writer,
            response: ServiceResponse,
    ) -> None:
        await self.service_response_codec.encode(writer, response)

    async def decode_topic_message_or_service_request(self, reader: Reader) -> TopicMessage | ServiceRequest:
        prefix = await reader.readexactly(1)

        if prefix == self.topic_message_prefix:
            return await self.topic_message_codec.decode(reader)
        elif prefix == self.service_request_prefix:
            return await self.service_request_codec.decode(reader)
        else:
            raise ValueError(f'Unknown prefix={prefix!r}')

    async def decode_service_response(self, reader: Reader) -> ServiceResponse:
        return await self.service_response_codec.decode(reader)



from argparse import ArgumentParser, Namespace
from pathlib import Path
from time import sleep

from rosy.cli.launch.args import ProcessArgs
from rosy.cli.launch.config import is_enabled, load_config
from rosy.procman import ProcessManager


async def launch_main(args: Namespace) -> None:
    _print(f"Using config: {args.config}")
    config = load_config(args.config)

    if args.exclude:
        _print(f"Excluding nodes: {args.exclude}")

    _print('Press Ctrl+C to stop all nodes.')

    with ProcessManager() as pm:
        node_args = start_coordinator(config, args.no_coordinator, pm)

        nodes = config['nodes']
        for node_name, node_config in nodes.items():
            if node_name in args.exclude:
                continue

            start_node(node_name, node_config, node_args, pm)

        try:
            pm.wait()
        except KeyboardInterrupt:
            pass


def add_launch_command(subparsers) -> None:
    parser: ArgumentParser = subparsers.add_parser(
        'launch',
        description="Launch rosy nodes together.",
        help="Launch rosy nodes together",
    )

    parser.add_argument(
        'config',
        nargs='?',
        default=Path('launch.yaml'),
        type=Path,
        help="Path to the configuration file. Default: %(default)s",
    )

    parser.add_argument(
        '--no-coordinator',
        action='store_true',
        help="Don't start the coordinator",
    )

    parser.add_argument(
        '-e', '--exclude',
        nargs='+',
        default=[],
        help='Nodes to exclude from starting',
    )


def start_coordinator(
        config: dict,
        disabled: bool,
        pm: ProcessManager,
) -> list[str]:
    """
    Start the coordinator (if not disabled), and return a list of coordinator
    arguments to pass to nodes.
    """

    node_args = []

    config = config.get('coordinator', {})

    args = ['rosy', 'coordinator']

    host = config.get('host')
    if host is not None:
        args.extend(['--host', host])

    port = config.get('port')
    if port is not None:
        args.extend(['--port', str(port)])

    client_host = config.get('client_host')
    if client_host or port:
        coordinator_arg = (client_host or '')
        if port:
            coordinator_arg += f':{port}'

        node_args.extend(['--coordinator', coordinator_arg])

    authkey = config.get('authkey')
    if authkey is not None:
        authkey_arg = ['--authkey', authkey]
        args.extend(authkey_arg)
        node_args.extend(authkey_arg)

    if not disabled and is_enabled(config):
        _print(f"Starting coordinator: {args}")
        pm.popen(args)

        delay = config.get('post_delay', 1)
        sleep(delay)
    else:
        _print('Not starting coordinator.')

    return node_args


def start_node(
        name: str,
        config: dict,
        coordinator_args: list[str],
        pm: ProcessManager,
) -> None:
    if not is_enabled(config):
        return

    delay = config.get('pre_delay', 0)
    sleep(delay)

    command = config['command']
    command = ProcessArgs(command)
    command.extend(['--name', name])
    command.extend(coordinator_args)
    command = command.args

    default_shell = isinstance(command, str)
    shell = config.get('shell', default_shell)

    number = config.get('number', 1)
    for i in range(number):
        _print(f'Starting node {name!r} ({i + 1}/{number}): {command}')
        pm.popen(command, shell=shell)

    delay = config.get('post_delay', 0)
    sleep(delay)


def _print(*args, **kwargs) -> None:
    print('[rosy launch]', *args, **kwargs)



from dataclasses import dataclass


@dataclass
class ProcessArgs:
    args: str | list[str]

    def append(self, arg: str) -> None:
        if isinstance(self.args, str):
            arg = quote_arg(arg)
            self.args = f'{self.args} {arg}'
        else:
            self.args.append(arg)

    def extend(self, args: list[str]) -> None:
        for arg in args:
            self.append(arg)


def quote_arg(arg: str) -> str:
    """Add double-quotes around an arg if necessary."""

    if not arg:
        return '""'
    elif ' ' in arg and not (arg[0] == arg[-1] == '"'):
        return f'"{arg}"'
    else:
        return arg
